/*
 GNU Maverik - a system for managing display and interaction in 
               Virtual Environment applications.
 Copyright (C) 1999-2002 Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - maverik@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/

#ifndef _MAV_COLLIDE_INCLUDE
#define _MAV_COLLIDE_INCLUDE

#ifdef MAVAPI
#undef MAVAPI
#endif
#if defined(WIN32) && !defined(__CYGWIN__) 
#ifdef LIBMAV_COLLIDE_EXPORTS
#define MAVAPI __declspec(dllexport) 
#else
#define MAVAPI __declspec(dllimport) 
#endif
#else
#define MAVAPI
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* Get point in object callback */

MAVAPI extern MAV_callback *mav_callback_getPoint;
typedef int (*MAV_callbackGetPointFn)(MAV_object *, MAV_vector *);
MAVAPI void mav_callbackGetPointSet(MAV_window *w, MAV_class *c, MAV_callbackGetPointFn fn);
MAVAPI int  mav_callbackGetPointExec(MAV_window *w, MAV_object *o, MAV_vector *v);

/* Get point(s) on plane callback */

MAVAPI extern MAV_callback *mav_callback_planePoint;
typedef int (*MAV_callbackPlanePointFn)(MAV_object *, MAV_line *, MAV_vector *);
MAVAPI void mav_callbackPlanePointSet(MAV_window *w, MAV_class *c, MAV_callbackPlanePointFn fn);
MAVAPI int  mav_callbackPlanePointExec(MAV_window *w, MAV_object *o, MAV_line plane, MAV_vector *v);

/* Get point callbacks for standard objects */

MAVAPI int mav_boxPlanePoint(MAV_object *obj, MAV_line *ln, MAV_vector *plane_point);
MAVAPI int mav_pyramidPlanePoint(MAV_object *obj, MAV_line *ln, MAV_vector *plane_point);
MAVAPI int mav_cylinderPlanePoint(MAV_object *obj, MAV_line *ln, MAV_vector *plane_point);
MAVAPI int mav_conePlanePoint(MAV_object *obj, MAV_line *ln, MAV_vector *plane_point);
MAVAPI int mav_spherePlanePoint(MAV_object *obj, MAV_line *ln, MAV_vector *plane_point);
MAVAPI int mav_hspherePlanePoint(MAV_object *obj, MAV_line *ln, MAV_vector *plane_point);
MAVAPI int mav_ellipsePlanePoint(MAV_object *obj, MAV_line *ln, MAV_vector *plane_point);
MAVAPI int mav_hellipsePlanePoint(MAV_object *obj, MAV_line *ln, MAV_vector *plane_point);
MAVAPI int mav_ctorusPlanePoint(MAV_object *obj, MAV_line *ln, MAV_vector *plane_point);
MAVAPI int mav_rtorusPlanePoint(MAV_object *obj, MAV_line *ln, MAV_vector *plane_point);
MAVAPI int mav_polygonPlanePoint(MAV_object *obj, MAV_line *ln, MAV_vector *plane_point);
MAVAPI int mav_facetPlanePoint(MAV_object *obj, MAV_line *ln, MAV_vector *plane_point);

/* Supporting routines */

#define MAV_COLLIDE_MAX_POINTS 100

MAVAPI int mav_collidePolygonObjectTest(MAV_polygon *, MAV_object *);
MAVAPI int mav_collidePolygonObjectPointGet(MAV_polygon *, MAV_object *, MAV_vector *);
MAVAPI int mav_collidePolygonObjectVectorTest(MAV_polygon *, MAV_object *, MAV_vector, float);
MAVAPI int mav_collidePolygonObjectVectorPointGet(MAV_polygon *, MAV_object *, MAV_vector, float, MAV_vector *);
MAVAPI int mav_collidePolygonCullVectorTest(MAV_polygon *, MAV_vector, float, MAV_list *);

MAVAPI int mav_BBPlanePoint(MAV_BB bb, MAV_line ln, MAV_vector *plane_point);



/* Module initialise */

MAVAPI char *mav_collideModuleID(void);
MAVAPI void mav_collideModuleInit(void);

#ifdef __cplusplus
}
#endif
#endif

