/*
 GNU Maverik - a system for managing display and interaction in 
               Virtual Environment applications.
 Copyright (C) 1999-2002 Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - maverik@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/

#ifndef _MAV_FFNAV_INCLUDE
#define _MAV_FFNAV_INCLUDE

#ifdef MAVAPI
#undef MAVAPI
#endif
#if defined(WIN32) && !defined(__CYGWIN__) 
#ifdef LIBMAV_FFNAV_EXPORTS
#define MAVAPI __declspec(dllexport) 
#else
#define MAVAPI __declspec(dllimport) 
#endif
#else
#define MAVAPI
#endif

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
  MAV_viewParams *vp;
  MAV_vector view;
  MAV_vector right;
  MAV_vector up;
  float height;
  float width;
  float depth;
  float eye_inset;
  float forehead;
  float knees;
  float shoesize;
  int falling;
} MAV_person;

MAVAPI extern float mav_navEpsilon;
MAVAPI extern float mav_navFallMax;
MAVAPI extern float mav_navClimbMax;
MAVAPI extern float mav_navMoveMax;
MAVAPI extern float mav_navMoveMin;

MAVAPI extern int mav_navPlaneXY;
MAVAPI extern int mav_navGravity;
MAVAPI extern int mav_navCollisions;
MAVAPI extern int mav_navFoot;
MAVAPI extern int mav_navForces;
MAVAPI extern int mav_navCollisionAccuracy;
MAVAPI extern int mav_navFootAccuracy;

MAVAPI extern float mav_navForcesZone;
MAVAPI extern float mav_navForcesGradient;
MAVAPI extern float mav_navForceRevMax;
MAVAPI extern float mav_navForceRotMax;

MAVAPI void mav_navInit(void);
MAVAPI void mav_navMove(MAV_person *, MAV_vector, float, MAV_list *);
MAVAPI void mav_navForcesMove(MAV_person *, MAV_vector, float, MAV_list *);
MAVAPI void mav_navDrop(MAV_person *, MAV_list *);
MAVAPI void mav_navRotate(MAV_person *, float, MAV_list *);

#include "mav_nearestPoint.h"

#ifdef __cplusplus
}
#endif
#endif

