/*
 GNU Maverik - a system for managing display and interaction in 
               Virtual Environment applications.
 Copyright (C) 1999-2002 Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - maverik@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/


#include "mavlib_collide.h"
#include <stdio.h>
#include <math.h>



/**************************************************************************/
/* mav_callback_getPoint  --- callback which returns a point on an object */
/**************************************************************************/

MAV_callback *mav_callback_getPoint;



/* Wrapper routines to set and execute get point callback */

void mav_callbackGetPointSet(MAV_window *w, MAV_class *c, MAV_callbackGetPointFn fn) 
{
  mav_callbackSet(mav_callback_getPoint, w, c, (MAV_callbackFn) fn);
}

int mav_callbackGetPointExec(MAV_window *w, MAV_object *o, MAV_vector *pt) 
{
  MAV_BB bb;
  int rv= 0;

  if (mav_callbackQuery(mav_callback_getPoint, w, o)) 
  {
    rv= mav_callbackExec(mav_callback_getPoint, w, o, pt, NULL);
  } 
  else if (mav_callbackQuery(mav_callback_BB, w, o)) 
  {
    if (mav_callbackBBExec(w, o, &bb)) {
      *pt= bb.min;
      rv= 1;
    }
  }

  return rv;
}



/* Callback functions for each object type */

/* Box */

int mav_boxGetPoint(MAV_object *obj, MAV_vector *pt) 
{
  MAV_box *box= (MAV_box *) mav_objectDataGet(obj);

  pt->x= 0;
  pt->y= 0;
  pt->z= 0;

  *pt= mav_vectorMult(*pt, box->matrix);

  return 1;
}



/* Pyramid */

int mav_pyramidGetPoint(MAV_object *obj, MAV_vector *pt) 
{
  MAV_pyramid *pyr= (MAV_pyramid *) mav_objectDataGet(obj);

  pt->x= 0;
  pt->y= 0;
  pt->z= 0;

  *pt= mav_vectorMult(*pt, pyr->matrix);

  return 1;
}



/* Cylinder */

int mav_cylinderGetPoint(MAV_object *obj, MAV_vector *pt) 
{
  MAV_cylinder *cyl= (MAV_cylinder *) mav_objectDataGet(obj);

  pt->x= 0;
  pt->y= 0;
  pt->z= 0;

  *pt= mav_vectorMult(*pt, cyl->matrix);

  return 1;
}



/* Cone */

int mav_coneGetPoint(MAV_object *obj, MAV_vector *pt) 
{
  MAV_cone *cone= (MAV_cone *) mav_objectDataGet(obj);

  pt->x= 0;
  pt->y= 0;
  pt->z= 0;

  *pt= mav_vectorMult(*pt, cone->matrix);

  return 1;
}



/* Sphere */

int mav_sphereGetPoint(MAV_object *obj, MAV_vector *pt) 
{
  MAV_sphere *sph= (MAV_sphere *) mav_objectDataGet(obj);

  pt->x= 0;
  pt->y= 0;
  pt->z= 0;

  *pt= mav_vectorMult(*pt, sph->matrix);

  return 1;
}



/* Half sphere */

int mav_hsphereGetPoint(MAV_object *obj, MAV_vector *pt) 
{
  MAV_hsphere *hsph= (MAV_hsphere *) mav_objectDataGet(obj);

  pt->x= 0;
  pt->y= 0;
  pt->z= 0;

  *pt= mav_vectorMult(*pt, hsph->matrix);

  return 1;
}



/* Ellipse */

int mav_ellipseGetPoint(MAV_object *obj, MAV_vector *pt) 
{
  MAV_ellipse *elip= (MAV_ellipse *) mav_objectDataGet(obj);

  pt->x= 0;
  pt->y= 0;
  pt->z= 0;

  *pt= mav_vectorMult(*pt, elip->matrix);

  return 1;
}



/* Half ellipse */

int mav_hellipseGetPoint(MAV_object *obj, MAV_vector *pt) 
{
  MAV_hellipse *helip= (MAV_hellipse *) mav_objectDataGet(obj);

  pt->x= 0;
  pt->y= 0;
  pt->z= 0;

  *pt= mav_vectorMult(*pt, helip->matrix);

  return 1;
}



/* Circular torus */

int mav_ctorusGetPoint(MAV_object *obj, MAV_vector *pt) 
{
  MAV_ctorus *ct= (MAV_ctorus *) mav_objectDataGet(obj);

  pt->x= ct->rmajor;
  pt->y= 0;
  pt->z= 0;

  *pt= mav_vectorMult(*pt, ct->matrix);

  return 1;
}




/* Rectangular torus */

int mav_rtorusGetPoint(MAV_object *obj, MAV_vector *pt) 
{
  MAV_rtorus *rt= (MAV_rtorus *) mav_objectDataGet(obj);

  pt->x= rt->radius;
  pt->y= 0;
  pt->z= 0;

  *pt= mav_vectorMult(*pt, rt->matrix);

  return 1;
}



/* Polygon */

int mav_polygonGetPoint(MAV_object *obj, MAV_vector *pt) 
{
  MAV_polygon *poly= (MAV_polygon *) mav_objectDataGet(obj);

  *pt= mav_vectorMult(poly->vert[0], poly->matrix);

  return 1;
}



/* Facet group */

int mav_facetGetPoint(MAV_object *obj, MAV_vector *pt) 
{
  MAV_facet *facet= (MAV_facet *) mav_objectDataGet(obj);

  *pt= mav_vectorMult(facet->vert[0][0], facet->matrix);

  return 1;
}



/* Routines to initialise the callbacks */

int mavlib_getPointInitialized= MAV_FALSE;

void mavlib_getPointInit(void) 
{
  if (!mavlib_getPointInitialized) {

    /* Define new callback */
    mav_callback_getPoint= mav_callbackNew();

    /* Set the callback for the default objects */
    mav_callbackGetPointSet(mav_win_all, mav_class_box, mav_boxGetPoint);
    mav_callbackGetPointSet(mav_win_all, mav_class_pyramid, mav_pyramidGetPoint);
    mav_callbackGetPointSet(mav_win_all, mav_class_cylinder, mav_cylinderGetPoint);
    mav_callbackGetPointSet(mav_win_all, mav_class_cone, mav_coneGetPoint);
    mav_callbackGetPointSet(mav_win_all, mav_class_sphere, mav_sphereGetPoint);
    mav_callbackGetPointSet(mav_win_all, mav_class_hsphere, mav_hsphereGetPoint);
    mav_callbackGetPointSet(mav_win_all, mav_class_ellipse, mav_ellipseGetPoint);
    mav_callbackGetPointSet(mav_win_all, mav_class_hellipse, mav_hellipseGetPoint);
    mav_callbackGetPointSet(mav_win_all, mav_class_ctorus, mav_ctorusGetPoint);
    mav_callbackGetPointSet(mav_win_all, mav_class_rtorus, mav_rtorusGetPoint);
    mav_callbackGetPointSet(mav_win_all, mav_class_polygon, mav_polygonGetPoint);
    mav_callbackGetPointSet(mav_win_all, mav_class_polygonGrp, NULL);
    mav_callbackGetPointSet(mav_win_all, mav_class_facet, mav_facetGetPoint);
    mav_callbackGetPointSet(mav_win_all, mav_class_rectangle, NULL);
    mav_callbackGetPointSet(mav_win_all, mav_class_polyline, NULL);
    mav_callbackGetPointSet(mav_win_all, mav_class_text, NULL);
    mav_callbackGetPointSet(mav_win_all, mav_class_composite, NULL);
    mav_callbackGetPointSet(mav_win_all, mav_class_SMSObj, NULL);
    mav_callbackGetPointSet(mav_win_all, mav_class_teapot, NULL);

    mavlib_getPointInitialized= MAV_TRUE;
  }
}
