/*
 GNU Maverik - a system for managing display and interaction in 
               Virtual Environment applications.
 Copyright (C) 1999-2002 Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - maverik@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/


#include "mavlib_collide.h"
#include <stdio.h>
#include <math.h>

/**************************************************************************/
/* mav_callback_planePoint - calculates (at least) one point per convex   */
/* area of a plane/object intersection                                    */
/**************************************************************************/

MAV_callback *mav_callback_planePoint;
int mavlib_planePointInitialized= MAV_FALSE;

/* BB plane test function */
int mav_BBPlanePoint (MAV_BB bb, MAV_line ln, MAV_vector *plane_point) {
/* only need to test three planes of bb */
  float sum;
  float x,y,z;
  float d;
  MAV_vector norm;

  d= mav_vectorDotProduct (ln.pt, ln.dir);
  norm= ln.dir;

/* first plane */
/* (z==min.z) */
  sum= d - norm.z * bb.min.z;

/* min.x */
  y= (sum - norm.x*bb.min.x)/norm.y;
  if ((y >= bb.min.y) && (y <= bb.max.y)) {
    plane_point->x= bb.min.x;
    plane_point->y= y;
    plane_point->z= bb.min.z;
    return 1;
  }

/* max.x */
  y= (sum - norm.x*bb.max.x)/norm.y;
  if ((y >= bb.min.y) && (y <= bb.max.y)) {
    plane_point->x= bb.max.x;
    plane_point->y= y;
    plane_point->z= bb.min.z;
    return 1;
  }

/* min.y */
  x= (sum - norm.y*bb.min.y)/norm.x;
  if ((x >= bb.min.x) && (x <= bb.max.x)) {
    plane_point->x= x;
    plane_point->y= bb.min.y;
    plane_point->z= bb.min.z;
    return 1;
  }

/* second plane */
/* (z==max.z) */
  sum= d - norm.z * bb.max.z;

/* min.x */
  y= (sum - norm.x*bb.min.x)/norm.y;
  if ((y >= bb.min.y) && (y <= bb.max.y)) {
    plane_point->x= bb.min.x;
    plane_point->y= y;
    plane_point->z= bb.max.z;
    return 1;
  }

/* max.x */
  y= (sum - norm.x*bb.max.x)/norm.y;
  if ((y >= bb.min.y) && (y <= bb.max.y)) {
    plane_point->x= bb.max.x;
    plane_point->y= y;
    plane_point->z= bb.max.z;
    return 1;
  }

/* min.y */
  x= (sum - norm.y*bb.min.y)/norm.x;
  if ((x >= bb.min.x) && (x <= bb.max.x)) {
    plane_point->x= x;
    plane_point->y= bb.min.y;
    plane_point->z= bb.max.z;
    return 1;
  }

/* third plane */
/* (x==max.x) */
  sum= d - norm.x * bb.max.x;

/* min.z */
  y= (sum - norm.z*bb.min.z)/norm.y;
  if ((y >= bb.min.y) && (y <= bb.max.y)) {
    plane_point->x= bb.max.x;
    plane_point->y= y;
    plane_point->z= bb.min.z;
    return 1;
  }

/* max.z */
  y= (sum - norm.z*bb.max.z)/norm.y;
  if ((y >= bb.min.y) && (y <= bb.max.y)) {
    plane_point->x= bb.max.x;
    plane_point->y= y;
    plane_point->z= bb.max.z;
    return 1;
  }

/* min.y */
  z= (sum - norm.y*bb.min.y)/norm.z;
  if ((z >= bb.min.z) && (z <= bb.max.z)) {
    plane_point->x= bb.max.x;
    plane_point->y= bb.min.y;
    plane_point->z= z;
    return 1;
  }

  return 0;
}

/* callback functions for each object type */

/* box */

int mav_boxPlanePoint (MAV_object *obj, MAV_line *ln, MAV_vector *plane_point)
{
  MAV_box *box= (MAV_box *) mav_objectDataGet (obj);
  MAV_line rot_ln;
  MAV_BB bb;
  MAV_vector pt;

  rot_ln= mav_lineTransFrame (*ln, box->matrix);
  bb.min= mav_vectorScalar (box->size, -0.5);
  bb.max= mav_vectorScalar (box->size, 0.5);

  if (mav_BBPlanePoint (bb, rot_ln, &pt)) {
    *plane_point= mav_vectorMult (pt, box->matrix);
    return 1;
  } else return 0;
}



/* cone */

int mav_conePlanePoint (MAV_object *obj, MAV_line *ln, MAV_vector *plane_point) {
  MAV_cone *cone= (MAV_cone *) mav_objectDataGet (obj);
  float dist1, dist2;
  MAV_line rot_ln;
  float soln;
  float thing;
  float test1, test2;

  rot_ln= mav_lineTransFrame (*ln, cone->matrix);
  soln= mav_vectorDotProduct (rot_ln.pt, rot_ln.dir);

  if ((rot_ln.dir.x==0) && (rot_ln.dir.y==0)) {
    thing= soln/rot_ln.dir.z;

    if ((thing >= -cone->height/2.0)&&(thing <= cone->height/2.0)) {
        plane_point->x= 0;
        plane_point->y= 0;
        plane_point->z= thing;
        *plane_point= mav_vectorMult (*plane_point, cone->matrix);
        return 1;
    } else return 0;
  } else {

    test1= cone->rt * sqrt (rot_ln.dir.x*rot_ln.dir.x+rot_ln.dir.y*rot_ln.dir.y);
    test2= cone->rb * sqrt (rot_ln.dir.x*rot_ln.dir.x+rot_ln.dir.y*rot_ln.dir.y);

    dist1= soln - rot_ln.dir.z * cone->height/2.0;
    dist2= soln + rot_ln.dir.z * cone->height/2.0;

    if (((dist1>test1)&&(dist2>test2))||
        ((dist1< -test1)&&(dist2< -test2)))
                return 0;

    if ((dist1<=test1)&&(dist1>= -test1)) {
        plane_point->x= dist1 * rot_ln.dir.x/ (rot_ln.dir.x*rot_ln.dir.x+rot_ln.dir.y*rot_ln.dir.y);
        plane_point->y= dist1 * rot_ln.dir.y/ (rot_ln.dir.x*rot_ln.dir.x+rot_ln.dir.y*rot_ln.dir.y);
        plane_point->z= cone->height/2.0;
        *plane_point= mav_vectorMult (*plane_point, cone->matrix);
        return 1;
    }

    if ((dist2<=test2)&&(dist2>= -test2)) {
        plane_point->x= dist2 * rot_ln.dir.x/ (rot_ln.dir.x*rot_ln.dir.x+rot_ln.dir.y*rot_ln.dir.y);
        plane_point->y= dist2 * rot_ln.dir.y/ (rot_ln.dir.x*rot_ln.dir.x+rot_ln.dir.y*rot_ln.dir.y);
        plane_point->z= -cone->height/2.0;
        *plane_point= mav_vectorMult (*plane_point, cone->matrix);
        return 1;
    }

    plane_point->x= 0;
    plane_point->y= 0;
    plane_point->z= soln/rot_ln.dir.z;
    *plane_point= mav_vectorMult (*plane_point, cone->matrix);
    return 1;
  }
}


/* circular torus */

int mav_ctorusPlanePoint (MAV_object *obj, MAV_line *ln, MAV_vector *plane_point) {
  MAV_line rot_ln;
  float height, soln;
  MAV_ctorus *ct= (MAV_ctorus *) mav_objectDataGet (obj);
  int num_pts= 0;
  float a, b, c, root, angle;
  float x, y, z, R;
  MAV_vector from, dir, offset;
  float length, dist;
  MAV_vector rot_pt, rot_norm;

  rot_ln= mav_lineTransFrame (*ln, ct->matrix);
  rot_pt= rot_ln.pt;
  rot_norm= rot_ln.dir;
  soln= mav_vectorDotProduct (rot_pt, rot_norm);

  if ((rot_norm.x==0)&&(rot_norm.y==0)) {
/* horizontal plane */
    height= soln/rot_norm.z;
    if ((height >= -ct->rminor) && (height <= ct->rminor)) {
        plane_point[num_pts].x= ct->rmajor;
        plane_point[num_pts].y= 0;
        plane_point[num_pts].z= height;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                ct->matrix);
        num_pts ++;
        return 1;
    } else return 0;
  }

  if (rot_norm.z==0) {
/* vertical plane */
/* check at z==0 */

/* outer circle */
    a= rot_norm.x*rot_norm.x + rot_norm.y*rot_norm.y;
    b= -2.0*soln*rot_norm.x;
    c= soln*soln - rot_norm.y*rot_norm.y*(ct->rmajor+ct->rminor)*
                                (ct->rmajor+ct->rminor);

    if (a != 0) {
      root= b*b - 4.0 * a*c;
      if (root > 0) {
        root= sqrt (root);
        x= (-b+root)/(2.0*a);
        y= (soln-rot_norm.x*x)/rot_norm.y;

        angle= atan2 (y, x);
        if (((angle>=0)&&(angle<=ct->angle))||((angle<0)&&
                ((angle+2.0*MAV_PI)<=ct->angle))) {
          plane_point[num_pts].x= x;
          plane_point[num_pts].y= y;
          plane_point[num_pts].z= 0;
          plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                ct->matrix);
          num_pts ++;
        }

        x= (-b-root)/(2.0*a);
        y= (soln-rot_norm.x*x)/rot_norm.y;

        angle= atan2 (y, x);
        if (((angle>=0)&&(angle<=ct->angle))||((angle<0)&&
                ((angle+2.0*MAV_PI)<=ct->angle))) {
          plane_point[num_pts].x= x;
          plane_point[num_pts].y= y;
          plane_point[num_pts].z= 0;
          plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                ct->matrix);
          num_pts ++;
        }

/* inner circle */
        c= soln*soln - rot_norm.y*rot_norm.y*(ct->rmajor-ct->rminor)*
                                (ct->rmajor-ct->rminor);
        root= b*b - 4.0 * a*c;
        if (root > 0) {
          root= sqrt (root);
          x= (-b+root)/(2.0*a);
          y= (soln-rot_norm.x*x)/rot_norm.y;

          angle= atan2 (y, x);
          if (((angle>=0)&&(angle<=ct->angle))||((angle<0)&&
                ((angle+2.0*MAV_PI)<=ct->angle))) {
            plane_point[num_pts].x= x;
            plane_point[num_pts].y= y;
            plane_point[num_pts].z= 0;
            plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                  ct->matrix);
            num_pts ++;
          }

          x= (-b-root)/(2.0*a);
          y= (soln-rot_norm.x*x)/rot_norm.y;

          angle= atan2 (y, x);
          if (((angle>=0)&&(angle<=ct->angle))||((angle<0)&&
                ((angle+2.0*MAV_PI)<=ct->angle))) {
            plane_point[num_pts].x= x;
            plane_point[num_pts].y= y;
            plane_point[num_pts].z= 0;
            plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                  ct->matrix);
            num_pts ++;
          }

        }

      }
    }

    return num_pts;
  }

/* circle at z==0 radius==rmajor */
/* only need this due to intersection calculation errors */
  a= rot_norm.x*rot_norm.x+rot_norm.y*rot_norm.y;
  b= -2.0*soln*rot_norm.x;
  c= soln*soln - rot_norm.y*rot_norm.y*ct->rmajor*ct->rmajor;

  if (a != 0) {
    root= b*b - 4.0*a*c;
    if (root > 0) {
      root= sqrt (root);
      x= (-b+root)/(2.0*a);
      y= (soln-rot_norm.x*x)/rot_norm.y;

      angle= atan2 (y,x);
      if (((angle>=0)&&(angle<=ct->angle))||((angle<0)&&
                ((angle+2.0*MAV_PI)<=ct->angle))) {
        plane_point[num_pts].x= x;
        plane_point[num_pts].y= y;
        plane_point[num_pts].z= 0;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                ct->matrix);
        num_pts ++;
      }

      x= (-b-root)/(2.0*a);
      y= (soln-rot_norm.x*x)/rot_norm.y;

      angle= atan2 (y,x);
      if (((angle>=0)&&(angle<=ct->angle))||((angle<0)&&
                ((angle+2.0*MAV_PI)<=ct->angle))) {
        plane_point[num_pts].x= x;
        plane_point[num_pts].y= y;
        plane_point[num_pts].z= 0;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                ct->matrix);
        num_pts ++;
      }
    }
  }

/* circle that would be touched by plane first */
  if (soln > 0) z= rot_norm.z*ct->rminor;
  else z= -rot_norm.z*ct->rminor;

  R= ct->rmajor + sqrt (ct->rminor*ct->rminor - z*z);

  a= rot_norm.x*rot_norm.x+rot_norm.y*rot_norm.y;
  b= 2.0*rot_norm.x*(rot_norm.z*z-soln);
  c= soln*soln + rot_norm.z*rot_norm.z*z*z - 2.0*rot_norm.z*z*soln -
                rot_norm.y*rot_norm.y*R*R;

  if (a != 0) {
    root= b*b - 4.0*a*c;
    if (root > 0) {
      root= sqrt (root);
      x= (-b+root)/(2.0*a);
      y= (soln-rot_norm.z*z-rot_norm.x*x)/rot_norm.y;

      angle= atan2 (y,x);
      if (((angle>=0)&&(angle<=ct->angle))||((angle<0)&&
                ((angle+2.0*MAV_PI)<=ct->angle))) {
        plane_point[num_pts].x= x;
        plane_point[num_pts].y= y;
        plane_point[num_pts].z= z;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                ct->matrix);
        num_pts ++;
      }

      x= (-b-root)/(2.0*a);
      y= (soln-rot_norm.z*z-rot_norm.x*x)/rot_norm.y;

      angle= atan2 (y,x);
      if (((angle>=0)&&(angle<=ct->angle))||((angle<0)&&
                ((angle+2.0*MAV_PI)<=ct->angle))) {
        plane_point[num_pts].x= x;
        plane_point[num_pts].y= y;
        plane_point[num_pts].z= z;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                ct->matrix);
        num_pts ++;
      }
    }
  }

/* first face */
  from.x= ct->rmajor-ct->rminor*rot_norm.x/sqrt (rot_norm.x*rot_norm.x+
                                rot_norm.z*rot_norm.z);
  from.y= 0;
  from.z= -ct->rminor*rot_norm.z/sqrt (rot_norm.x*rot_norm.x+
                rot_norm.z*rot_norm.z);
  dir.x= 2.0*(ct->rmajor-from.x);
  dir.y= 0;
  dir.z= -2.0*from.z;
  length= sqrt (mav_vectorDotProduct (dir, dir));
  dir= mav_vectorNormalize (dir);

  if (mav_vectorDotProduct (rot_norm, dir) != 0) {
    dist= (soln - mav_vectorDotProduct (rot_norm, from))/mav_vectorDotProduct (rot_norm, dir);
    if ((dist > -MAV_EPSILON)&&(dist<=length)) {
        plane_point[num_pts].x= from.x+dist*dir.x;
        plane_point[num_pts].y= 0;
        plane_point[num_pts].z= from.z+dist*dir.z;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                ct->matrix);
        num_pts ++;
    }
  }

/* second face */
  offset.x= cos (ct->angle) * (cos (ct->angle) * rot_norm.x +
                        sin (ct->angle) * rot_norm.y);;
  offset.y= sin (ct->angle) * (cos (ct->angle) * rot_norm.y +
                        sin (ct->angle) * rot_norm.y);
  offset.z= rot_norm.z;
  offset= mav_vectorNormalize (offset);

  from.x= cos (ct->angle) * ct->rmajor - ct->rminor*offset.x;
  from.y= sin (ct->angle) * ct->rmajor - ct->rminor*offset.y;
  from.z= -ct->rminor *offset.z;

  dir.x= 2.0* offset.x*ct->rminor;
  dir.y= 2.0* offset.y*ct->rminor;
  dir.z= -2.0*from.z;

  length= sqrt (mav_vectorDotProduct (dir, dir));
  dir= mav_vectorNormalize (dir);

  if (mav_vectorDotProduct (rot_norm, dir) != 0) {
    dist= (soln - mav_vectorDotProduct (rot_norm, from))/mav_vectorDotProduct (rot_norm, dir);
    if ((dist > -MAV_EPSILON)&&(dist<=length)) {
        plane_point[num_pts].x= from.x+dist*dir.x;
        plane_point[num_pts].y= from.y+dist*dir.y;
        plane_point[num_pts].z= from.z+dist*dir.z;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                ct->matrix);
        num_pts ++;
    }
  }

  return num_pts;
}


/* cylinder */

int mav_cylinderPlanePoint (MAV_object *obj, MAV_line *ln, MAV_vector *plane_point) {
  MAV_cylinder *cyl= (MAV_cylinder *) mav_objectDataGet (obj);
  float dist1, dist2;
  float thing;
  float soln;
  float tester;
  MAV_vector rot_pt, rot_norm;
  MAV_line rot_ln;

  rot_ln= mav_lineTransFrame (*ln, cyl->matrix);
  rot_pt= rot_ln.pt;
  rot_norm= rot_ln.dir;
  soln= mav_vectorDotProduct (rot_pt, rot_norm);

  if ((rot_norm.x==0)&&(rot_norm.y==0)) {
    thing= soln/rot_norm.z;

    if ((thing >= -cyl->height/2.0)&&(thing <= cyl->height/2.0)) {
        plane_point->x= 0;
        plane_point->y= 0;
        plane_point->z= thing;
        *plane_point= mav_vectorMult (*plane_point, cyl->matrix);
        return 1;
    } else return 0;
  } else {

    if (rot_norm.z > 0.0) {
      dist1= soln - rot_norm.z * cyl->height/2.0;
      dist2= soln + rot_norm.z * cyl->height/2.0;
    } else {
      dist1= soln + rot_norm.z * cyl->height/2.0;
      dist2= soln - rot_norm.z * cyl->height/2.0;
    }

    tester= cyl->radius*sqrt (rot_norm.x*rot_norm.x + rot_norm.y*rot_norm.y);

    if ((dist1>tester)||(dist2< -tester))
        return 0;

    if (dist1>= -tester) {
        plane_point->x= dist1 * rot_norm.x / (rot_norm.x*rot_norm.x+rot_norm.y*rot_norm.y);
        plane_point->y= dist1 * rot_norm.y / (rot_norm.x*rot_norm.x+rot_norm.y*rot_norm.y);
        if (rot_norm.z > 0.0)
          plane_point->z= cyl->height/2.0;
        else
          plane_point->z= -cyl->height/2.0;
        *plane_point= mav_vectorMult (*plane_point, cyl->matrix);
        return 1;
    }

    if (dist2 <= tester) {
        plane_point->x= dist2 * rot_norm.x / (rot_norm.x*rot_norm.x+rot_norm.y*rot_norm.y);
        plane_point->y= dist2 * rot_norm.y / (rot_norm.x*rot_norm.x+rot_norm.y*rot_norm.y);
        if (rot_norm.z > 0.0)
          plane_point->z= -cyl->height/2.0;
        else
          plane_point->z= cyl->height/2.0;
        *plane_point= mav_vectorMult (*plane_point, cyl->matrix);
        return 1;
    }

    plane_point->x= 0;
    plane_point->y= 0;
    plane_point->z= soln/rot_norm.z;
    *plane_point= mav_vectorMult (*plane_point, cyl->matrix);
    return 1;
  }
}


/* ellipse */

int mav_ellipsePlanePoint (MAV_object *obj, MAV_line *ln, MAV_vector *plane_point) {
  MAV_ellipse *elip= (MAV_ellipse *) mav_objectDataGet (obj);
  float thing, a, b, c, soln, z1, z2;
  MAV_vector rot_pt, rot_norm;
  MAV_line rot_ln;

  rot_ln= mav_lineTransFrame (*ln, elip->matrix);
  rot_pt= rot_ln.pt;
  rot_norm= rot_ln.dir;
  soln= mav_vectorDotProduct (rot_pt, rot_norm);

  if ((rot_norm.x==0)&&(rot_norm.y==0)) {
    thing= soln/rot_norm.z;

    if ((thing >= -elip->height)&&(thing <= elip->height)) {
        plane_point->x= 0;
        plane_point->y= 0;
        plane_point->z= thing;
        *plane_point= mav_vectorMult (*plane_point, elip->matrix);
        return 1;
    } else return 0;
  } else {

        a= rot_norm.z*rot_norm.z + elip->radius*elip->radius * (rot_norm.x*rot_norm.x +
                rot_norm.y*rot_norm.y) / (elip->height * elip->height);
        b= -2.0 * soln * rot_norm.z;
        c= soln*soln - elip->radius*elip->radius * (rot_norm.x*rot_norm.x +
                rot_norm.y*rot_norm.y);

        if (a==0) return 0;

        thing= b*b - 4.0 *a*c;
        if (thing<=0) return 0;
        thing= sqrt (thing);

        z1= (-b+thing)/(2.0*a);
        z2= (-b-thing)/(2.0*a);

        if ((z1>= -elip->height)&&(z1<=elip->height)) {
                plane_point->x= rot_norm.x * (soln - z1 * rot_norm.z) /
                        sqrt (rot_norm.x*rot_norm.x+rot_norm.y*rot_norm.y);
                plane_point->y= rot_norm.y * (soln - z1 * rot_norm.z) /
                        sqrt (rot_norm.x*rot_norm.x+rot_norm.y*rot_norm.y);
                plane_point->z= z1;
                *plane_point= mav_vectorMult (*plane_point, elip->matrix);
                return 1;
        } else if ((z2>= -elip->height)&&(z2<=elip->height)) {
                plane_point->x= rot_norm.x * (soln - z2 * rot_norm.z) /
                        sqrt (rot_norm.x*rot_norm.x+rot_norm.y*rot_norm.y);
                plane_point->y= rot_norm.y * (soln - z2 * rot_norm.z) /
                        sqrt (rot_norm.x*rot_norm.x+rot_norm.y*rot_norm.y);
                plane_point->z= z2;
                *plane_point= mav_vectorMult (*plane_point, elip->matrix);
                return 1;
        }

  }

  return 0;
}

/* facet group */

int mav_facetPlanePoint (MAV_object *obj, MAV_line *ln, MAV_vector *plane_point) {
  MAV_facet *facet= (MAV_facet *) mav_objectDataGet (obj);
  int i, count;
  MAV_vector this, next, edge;
  float length, dist;
  float soln;
  MAV_vector rot_norm, rot_pt;
  MAV_line rot_ln;
  int num_pts= 0;

  rot_ln= mav_lineTransFrame (*ln, facet->matrix);
  rot_pt= rot_ln.pt;
  rot_norm= rot_ln.dir;
  soln= mav_vectorDotProduct (rot_pt, rot_norm);

  for (i=0; i<facet->npolys; i++) {
    for (count= 0; count < (facet->np[i]-1); count++) {
        this= facet->vert[i][count];
        next= facet->vert[i][count+1];

        dist= -1.0;

        edge= mav_vectorSub (next, this);
        length= mav_vectorDotProduct (edge, edge);
        if (length>0) length= sqrt (length);
        edge= mav_vectorNormalize (edge);

        if (mav_vectorDotProduct (rot_norm, edge) != 0) {
          dist= (soln - mav_vectorDotProduct (rot_norm, this))/
                mav_vectorDotProduct (rot_norm, edge);
          if ((dist > -MAV_EPSILON) && (dist <= length)) {
	    *plane_point= mav_vectorAdd (this,
		mav_vectorScalar (edge, dist));
            *plane_point= mav_vectorMult (*plane_point, facet->matrix);
	    if (num_pts<(MAV_COLLIDE_MAX_POINTS-1)) {
	      plane_point ++;
              num_pts ++;
	    } else fprintf (stderr, "Collide lib vertex overflow\n");
          }
        }
    }
  }

  return num_pts;
}

/* half ellipse */

int mav_hellipsePlanePoint (MAV_object *obj, MAV_line *ln, MAV_vector *plane_point) {
  MAV_hellipse *helip= (MAV_hellipse *) mav_objectDataGet (obj);
  float thing, a, b, c, soln, z1, z2;
  MAV_vector rot_pt, rot_norm;
  MAV_line rot_ln;

  rot_ln= mav_lineTransFrame (*ln, helip->matrix);
  rot_pt= rot_ln.pt;
  rot_norm= rot_ln.dir;
  soln= mav_vectorDotProduct (rot_pt, rot_norm);

  if ((rot_norm.x==0) && (rot_norm.y==0)) {
    thing= soln/rot_norm.z;

    if ((thing >= 0) && (thing <= helip->height)) {
        plane_point->x= 0;
        plane_point->y= 0;
        plane_point->z= thing;
        *plane_point= mav_vectorMult (*plane_point, helip->matrix);
        return 1;
    } else return 0;
  } else {

        a= rot_norm.z*rot_norm.z + helip->radius*helip->radius *
                (rot_norm.x*rot_norm.x + rot_norm.y*rot_norm.y) /
                (helip->height * helip->height);
        b= -2.0 * soln * rot_norm.z;
        c= soln*soln - helip->radius*helip->radius *
                (rot_norm.x*rot_norm.x + rot_norm.y*rot_norm.y);

        if (a==0) return 0;

        thing= b*b - 4.0 * a*c;
        if (thing<=0) return 0;
        thing= sqrt(thing);

        z1= (-b+thing)/(2.0*a);
        z2= (-b-thing)/(2.0*a);

        if ((z1 >= 0) && (z1 <= helip->height)) {
                plane_point->x= rot_norm.x * (soln - z1 * rot_norm.z) /
                        sqrt (rot_norm.x*rot_norm.x + rot_norm.y*rot_norm.y);
                plane_point->y= rot_norm.y * (soln - z1 * rot_norm.z) /
                        sqrt (rot_norm.x*rot_norm.x + rot_norm.y*rot_norm.y);
                plane_point->z= z1;
                *plane_point= mav_vectorMult (*plane_point, helip->matrix);
                return 1;
        } else if ((z2 >= 0) && (z2 <= helip->height)) {
                plane_point->x= rot_norm.x * (soln - z2 * rot_norm.z) /
                        sqrt (rot_norm.x*rot_norm.x+rot_norm.y*rot_norm.y);
                plane_point->y= rot_norm.y * (soln - z2 * rot_norm.z) /
                        sqrt (rot_norm.x*rot_norm.x+rot_norm.y*rot_norm.y);
                plane_point->z= z2;
                *plane_point= mav_vectorMult (*plane_point, helip->matrix);
                return 1;
        }

  }

  return 0;
}

/* half sphere */
int mav_hspherePlanePoint (MAV_object *obj, MAV_line *ln, MAV_vector *plane_point) {
  MAV_hsphere *hsph= (MAV_hsphere *) mav_objectDataGet (obj);
  float dist, soln, thing;
  MAV_vector rot_norm, rot_pt;
  MAV_line rot_ln;

  rot_ln= mav_lineTransFrame (*ln, hsph->matrix);
  rot_pt= rot_ln.pt;
  rot_norm= rot_ln.dir;
  soln= mav_vectorDotProduct (rot_pt, rot_norm);

  dist= soln;

  if ((dist >= -hsph->radius) && (dist <= hsph->radius) &&
        ((dist*rot_norm.z) >= 0)) {
    *plane_point= mav_vectorMult (mav_vectorScalar (rot_norm, dist),
		hsph->matrix);
    return 1;
  }

  thing= rot_norm.x*rot_norm.x + rot_norm.y*rot_norm.y;

  if (thing > 0) {
    dist= soln / sqrt (thing);
    if ((dist >= -hsph->radius) && (dist <= hsph->radius)) {
      plane_point->x= soln*rot_norm.x/
                (rot_norm.x*rot_norm.x + rot_norm.y*rot_norm.y);
      plane_point->y= soln*rot_norm.y/
                (rot_norm.x*rot_norm.x + rot_norm.y*rot_norm.y);
      plane_point->z= 0;
      *plane_point= mav_vectorMult (*plane_point, hsph->matrix);
      return 1;
    }
  }

  return 0;
}

/* polygon */

int mav_polygonPlanePoint (MAV_object *obj, MAV_line *ln, MAV_vector *plane_point) {
  MAV_polygon *poly= (MAV_polygon *) mav_objectDataGet (obj);
  int count;
  MAV_vector this, next, edge;
  float length, dist;
  float soln;
  MAV_line rot_ln;
  int use_this_one= 1;
  int num_pts= 0;

  rot_ln= mav_lineTransFrame (*ln, poly->matrix);
  soln= mav_vectorDotProduct (rot_ln.pt, rot_ln.dir);

  for (count= 0; count < (poly->np-1); count ++) {
    this= poly->vert[count];
    next= poly->vert[count+1];

    edge= mav_vectorSub (next, this);
    length= mav_vectorDotProduct (edge, edge);
    if (length>0) length= sqrt (length);
    edge= mav_vectorNormalize (edge);

    if (mav_vectorDotProduct (rot_ln.dir, edge) != 0) {
      dist= (soln - mav_vectorDotProduct (rot_ln.dir, this)) /
                mav_vectorDotProduct (rot_ln.dir, edge);

      if ((dist > -MAV_EPSILON) && (dist <=length)) {
        if (use_this_one) {
	  *plane_point= mav_vectorAdd (this,
		mav_vectorScalar (edge, dist));
          *plane_point= mav_vectorMult (*plane_point, poly->matrix);
	  if (num_pts<(MAV_COLLIDE_MAX_POINTS-1)) {
            num_pts ++;
	    plane_point ++;
	  } else fprintf (stderr, "Collide lib vertex overflow\n");
        }
        use_this_one= !use_this_one;
      }
    }
  }

  return num_pts;
}


/* pyramid */

int mav_pyramidPlanePoint (MAV_object *obj, MAV_line *ln, MAV_vector *plane_point) {
  MAV_pyramid *pyr= (MAV_pyramid *) mav_objectDataGet (obj);
  MAV_vector rot_norm, rot_pt;
  float soln, dist, length;
  float thing, x, y;
  MAV_vector vertex, dir, temp;
  MAV_line rot_ln;
  float bot_xmin, bot_xmax, bot_ymin, bot_ymax, bot_z;
  float top_xmin, top_xmax, top_ymin, top_ymax, top_z;

  bot_xmin = -pyr->offset_x/2.0-pyr->bot_size_x/2.0;
  bot_xmax = -pyr->offset_x/2.0+pyr->bot_size_x/2.0;
  bot_ymin = -pyr->offset_y/2.0-pyr->bot_size_y/2.0;
  bot_ymax = -pyr->offset_y/2.0+pyr->bot_size_y/2.0;
  bot_z = -pyr->height/2.0;

  top_xmin = pyr->offset_x/2.0-pyr->top_size_x/2.0;
  top_xmax = pyr->offset_x/2.0+pyr->top_size_x/2.0;
  top_ymin = pyr->offset_y/2.0-pyr->top_size_y/2.0;
  top_ymax = pyr->offset_y/2.0+pyr->top_size_y/2.0;
  top_z = pyr->height/2.0;


  rot_ln= mav_lineTransFrame (*ln, pyr->matrix);
  rot_pt= rot_ln.pt;
  rot_norm= rot_ln.dir;
  soln= mav_vectorDotProduct (rot_pt, rot_norm);

/* test top face, bottom face and a line between them */

/* bottom face (z==bot_z) */

  thing= soln - rot_norm.z * bot_z;

/* min.x */
  y= (thing - rot_norm.x * bot_xmin)/rot_norm.y;
  if ((y >= bot_ymin) && (y <= bot_ymax)) {
    plane_point->x= bot_xmin;
    plane_point->y= y;
    plane_point->z= bot_z;
    *plane_point= mav_vectorMult (*plane_point, pyr->matrix);
    return 1;
  }

/* max.x */
  y= (thing - rot_norm.x * bot_xmax)/rot_norm.y;
  if ((y >= bot_ymin) && (y <= bot_ymax)) {
    plane_point->x= bot_xmax;
    plane_point->y= y;
    plane_point->z= bot_z;
    *plane_point= mav_vectorMult (*plane_point, pyr->matrix);
    return 1;
  }

/* min.y */
  x= (thing - rot_norm.y * bot_ymin)/rot_norm.x;
  if ((x >= bot_xmin) && (x <= bot_xmax)) {
    plane_point->x= x;
    plane_point->y= bot_ymin;
    plane_point->z= bot_z;
    *plane_point= mav_vectorMult (*plane_point, pyr->matrix);
    return 1;
  }

/* top face (z==top_z) */

  thing= soln - rot_norm.z * top_z;

/* min.x */
  y= (thing - rot_norm.x * top_xmin)/rot_norm.y;
  if ((y >= top_ymin) && (y <= top_ymax)) {
    plane_point->x= top_xmin;
    plane_point->y= y;
    plane_point->z= top_z;
    *plane_point= mav_vectorMult (*plane_point, pyr->matrix);
    return 1;
  }

/* max.x */
  y= (thing - rot_norm.x * top_xmax)/rot_norm.y;
  if ((y >= top_ymin) && (y <= top_ymax)) {
    plane_point->x= top_xmax;
    plane_point->y= y;
    plane_point->z= top_z;
    *plane_point= mav_vectorMult (*plane_point, pyr->matrix);
    return 1;
  }

/* min.y */
  x= (thing - rot_norm.y * top_ymin)/rot_norm.x;
  if ((x >= top_xmin) && (x <= top_xmax)) {
    plane_point->x= x;
    plane_point->y= top_ymin;
    plane_point->z= top_z;
    *plane_point= mav_vectorMult (*plane_point, pyr->matrix);
    return 1;
  }

/* line between botmin and topmin */
  vertex.x= bot_xmin;
  vertex.y= bot_ymin;
  vertex.z= bot_z;

  temp.x= top_xmin;
  temp.y= top_ymin;
  temp.z= top_z;

  dir= mav_vectorSub (temp, vertex);
  length= mav_vectorDotProduct (dir, dir);
  if (length> 0) length= sqrt (length);
  dir= mav_vectorNormalize (dir);

  if (mav_vectorDotProduct (rot_norm, dir) != 0) {
    dist= (soln - mav_vectorDotProduct (rot_norm, vertex))/mav_vectorDotProduct (rot_norm, dir);
    if ((dist > -MAV_EPSILON) && (dist <= length)) {
	*plane_point= mav_vectorAdd (vertex,
		mav_vectorScalar (dir, dist));
        *plane_point= mav_vectorMult (*plane_point, pyr->matrix);
        return 1;
    }
  }

  return 0;
}

/* rectangular torus */

int mav_rtorusPlanePoint (MAV_object *obj, MAV_line *ln, MAV_vector *plane_point) {
  MAV_rtorus *rt= (MAV_rtorus *) mav_objectDataGet (obj);
  MAV_vector rot_pt, rot_norm;
  float soln;
  float root, thing, x, y, z;
  float a, b, c;
  int num_pts= 0;
  MAV_line rot_ln;

  rot_ln= mav_lineTransFrame (*ln, rt->matrix);
  rot_pt= rot_ln.pt;
  rot_norm= rot_ln.dir;
  soln= mav_vectorDotProduct (rot_pt, rot_norm);

/* test each face and top and bottom surfaces */

/* top surface */

  if ((rot_norm.x==0) && (rot_norm.y==0)) {
    thing= soln/rot_norm.z;
    if ((thing >= -rt->height/2.0) && (thing <= rt->height/2.0)) {
        plane_point[num_pts].x= rt->radius-rt->width/2.0;
        plane_point[num_pts].y= 0;
        plane_point[num_pts].z= thing;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                rt->matrix);
        num_pts ++;
        return 1;
    } else return 0;
  }

  a= rot_norm.x*rot_norm.x + rot_norm.y*rot_norm.y;
  b= rot_norm.y * (rot_norm.z*rt->height - 2.0 * soln);
  c= soln*soln + rt->height * rt->height * rot_norm.z*rot_norm.z /4.0 - rt->height * soln * rot_norm.z -
                rot_norm.x*rot_norm.x * (rt->radius+rt->width/2.0) *
                                (rt->radius+rt->width/2.0);

  if (a!=0) {
    root= b*b - 4.0*a*c;
    if (root > 0) {
      root= sqrt (root);
      y= (-b+root)/(2.0*a);
      x= (soln-rot_norm.z*rt->height/2.0-rot_norm.y*y)/rot_norm.x;

      thing= atan2 (y,x);
      if (((thing>=0)&&(thing<=rt->angle))||((thing<0)&&
                ((thing+2.0*MAV_PI)<=rt->angle))) {
        plane_point[num_pts].x= x;
        plane_point[num_pts].y= y;
        plane_point[num_pts].z= rt->height/2.0;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                rt->matrix);
        num_pts ++;
      }

      y= (-b-root)/(2.0*a);
      x= (soln-rot_norm.z*rt->height/2.0-rot_norm.y*y)/rot_norm.x;

      thing= atan2 (y,x);
      if (((thing>=0)&&(thing<=rt->angle))||((thing<0)&&
                ((thing+2.0*MAV_PI)<=rt->angle))) {
        plane_point[num_pts].x= x;
        plane_point[num_pts].y= y;
        plane_point[num_pts].z= rt->height/2.0;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                rt->matrix);
        num_pts ++;
      }

    }
  }

  a= rot_norm.x*rot_norm.x + rot_norm.y*rot_norm.y;
  b= rot_norm.y * (rot_norm.z*rt->height - 2.0 * soln);
  c= soln*soln + rt->height * rt->height *rot_norm.z*rot_norm.z /4.0 - rt->height * soln *rot_norm.z -
                rot_norm.x*rot_norm.x * (rt->radius-rt->width/2.0) *
                                (rt->radius-rt->width/2.0);

  if (a!=0) {
    root= b*b - 4.0*a*c;
    if (root > 0) {
      root= sqrt (root);
      y= (-b+root)/(2.0*a);
      x= (soln-rot_norm.z*rt->height/2.0-rot_norm.y*y)/rot_norm.x;
        
      thing= atan2 (y,x);
      if (((thing>=0)&&(thing<=rt->angle))||((thing<0)&&
                ((thing+2.0*MAV_PI)<=rt->angle))) {
        plane_point[num_pts].x= x;
        plane_point[num_pts].y= y;
        plane_point[num_pts].z= rt->height/2.0;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                rt->matrix);
        num_pts ++;
      }

      y= (-b-root)/(2.0*a);
      x= (soln-rot_norm.z*rt->height/2.0-rot_norm.y*y)/rot_norm.x;

      thing= atan2 (y,x);
      if (((thing>=0)&&(thing<=rt->angle))||((thing<0)&&
                ((thing+2.0*MAV_PI)<=rt->angle))) {
        plane_point[num_pts].x= x;
        plane_point[num_pts].y= y;
        plane_point[num_pts].z= rt->height/2.0;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                rt->matrix);
        num_pts ++;
      }

    }
  }

/* bottom surface */
  a= rot_norm.x*rot_norm.x + rot_norm.y*rot_norm.y;
  b= rot_norm.y * (-rot_norm.z*rt->height - 2.0 * soln);
  c= soln*soln + rt->height * rt->height *rot_norm.z*rot_norm.z /4.0 + rt->height * soln *rot_norm.z -
                rot_norm.x*rot_norm.x * (rt->radius+rt->width/2.0) *
                                (rt->radius+rt->width/2.0);

  if (a!=0) {
    root= b*b - 4.0*a*c;
    if (root > 0) {
      root= sqrt (root);
      y= (-b+root)/(2.0*a);
      x= (soln+rot_norm.z*rt->height/2.0-rot_norm.y*y)/rot_norm.x;

      thing= atan2 (y,x);
      if (((thing>=0)&&(thing<=rt->angle))||((thing<0)&&
                ((thing+2.0*MAV_PI)<=rt->angle))) {
        plane_point[num_pts].x= x;
        plane_point[num_pts].y= y;
        plane_point[num_pts].z= -rt->height/2.0;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                rt->matrix);
        num_pts ++;
      }

      y= (-b-root)/(2.0*a);
      x= (soln+rot_norm.z*rt->height/2.0-rot_norm.y*y)/rot_norm.x;

      thing= atan2 (y,x);
      if (((thing>=0)&&(thing<=rt->angle))||((thing<0)&&
                ((thing+2.0*MAV_PI)<=rt->angle))) {
        plane_point[num_pts].x= x;
        plane_point[num_pts].y= y;
        plane_point[num_pts].z= -rt->height/2.0;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                rt->matrix);
        num_pts ++;
      }

    }
  }

  a= rot_norm.x*rot_norm.x + rot_norm.y*rot_norm.y;
  b= rot_norm.y * (-rot_norm.z*rt->height - 2.0 * soln);
  c= soln*soln + rt->height * rt->height*rot_norm.z*rot_norm.z /4.0 + rt->height * soln*rot_norm.z -
                rot_norm.x*rot_norm.x * (rt->radius-rt->width/2.0) *
                                (rt->radius-rt->width/2.0);

  if (a!=0) {
    root= b*b - 4.0*a*c;
    if (root > 0) {
      root= sqrt (root);
      y= (-b+root)/(2.0*a);
      x= (soln+rot_norm.z*rt->height/2.0-rot_norm.y*y)/rot_norm.x;
        
      thing= atan2 (y,x);
      if (((thing>=0)&&(thing<=rt->angle))||((thing<0)&&
                ((thing+2.0*MAV_PI)<=rt->angle))) {
        plane_point[num_pts].x= x;
        plane_point[num_pts].y= y;
        plane_point[num_pts].z= -rt->height/2.0;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                rt->matrix);
        num_pts ++;
      }

      y= (-b-root)/(2.0*a);
      x= (soln+rot_norm.z*rt->height/2.0-rot_norm.y*y)/rot_norm.x;

      thing= atan2 (y,x);
      if (((thing>=0)&&(thing<=rt->angle))||((thing<0)&&
                ((thing+2.0*MAV_PI)<=rt->angle))) {
        plane_point[num_pts].x= x;
        plane_point[num_pts].y= y;
        plane_point[num_pts].z= -rt->height/2.0;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                rt->matrix);
        num_pts ++;
      }
    }
  }

/* face 1 */
/* y==0 */
  thing= soln;

/* x==radius-width/2 */
  z= (thing - rot_norm.x*(rt->radius - rt->width/2.0))/rot_norm.z;
  if ((z >= -rt->height/2.0) && (z <= rt->height/2.0)) {
        plane_point[num_pts].x= rt->radius - rt->width/2.0;
        plane_point[num_pts].y= 0;
        plane_point[num_pts].z= z;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                        rt->matrix);
        num_pts ++;
  } else {

/* x==radius+width/2 */
    z= (thing - rot_norm.x*(rt->radius + rt->width/2.0))/rot_norm.z;
    if ((z >= -rt->height/2.0) && (z <= rt->height/2.0)) {
        plane_point[num_pts].x= rt->radius + rt->width/2.0;
        plane_point[num_pts].y= 0;
        plane_point[num_pts].z= z;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                        rt->matrix);
        num_pts ++;
    } else {

/*  z==height/2.0 */
      x= (thing - rot_norm.z*rt->height/2.0)/rot_norm.x;
      if ((x >= (rt->radius-rt->width/2.0)) && (x <= (rt->radius+rt->width/2.0))) {
        plane_point[num_pts].x= x;
        plane_point[num_pts].y= 0;
        plane_point[num_pts].z= rt->height/2.0;
        plane_point[num_pts]= mav_vectorMult (plane_point[num_pts],
                        rt->matrix);
        num_pts ++;
      }
    }
  }

  return num_pts;
}

/* sphere */

int mav_spherePlanePoint (MAV_object *obj, MAV_line *ln, MAV_vector *plane_point) {
  MAV_sphere *sph= (MAV_sphere *) mav_objectDataGet (obj);
  float dist, soln;
  MAV_vector rot_norm, rot_pt;
  MAV_line rot_ln;

  rot_ln= mav_lineTransFrame (*ln, sph->matrix);
  rot_pt= rot_ln.pt;
  rot_norm= rot_ln.dir;
  soln= mav_vectorDotProduct (rot_pt, rot_norm);

  dist= soln;

  if ((dist >= -sph->radius) && (dist <= sph->radius)) {
    *plane_point= mav_vectorScalar (rot_norm, dist);
    *plane_point= mav_vectorMult (*plane_point, sph->matrix);
    return 1;
  }

  return 0;
}


/*******************************************************************/
/*******************************************************************/


void mav_planePointInit (void) {
  if (!mavlib_planePointInitialized) {
    mav_callback_planePoint= mav_callbackNew ();

    mav_callbackPlanePointSet (mav_win_all, mav_class_cone, mav_conePlanePoint);
    mav_callbackPlanePointSet (mav_win_all, mav_class_ctorus, mav_ctorusPlanePoint);
    mav_callbackPlanePointSet (mav_win_all, mav_class_cylinder, mav_cylinderPlanePoint);
    mav_callbackPlanePointSet (mav_win_all, mav_class_ellipse, mav_ellipsePlanePoint);
    mav_callbackPlanePointSet (mav_win_all, mav_class_facet, mav_facetPlanePoint);
    mav_callbackPlanePointSet (mav_win_all, mav_class_hellipse, mav_hellipsePlanePoint);
    mav_callbackPlanePointSet (mav_win_all, mav_class_hsphere, mav_hspherePlanePoint);
    mav_callbackPlanePointSet (mav_win_all, mav_class_polygon, mav_polygonPlanePoint);
    mav_callbackPlanePointSet (mav_win_all, mav_class_pyramid, mav_pyramidPlanePoint);
    mav_callbackPlanePointSet (mav_win_all, mav_class_rtorus, mav_rtorusPlanePoint);
    mav_callbackPlanePointSet (mav_win_all, mav_class_sphere, mav_spherePlanePoint);
    mav_callbackPlanePointSet (mav_win_all, mav_class_box, mav_boxPlanePoint);
    mavlib_planePointInitialized= MAV_TRUE;
  }
}

/* wrapper routines to set and execute plane point callback */

void mav_callbackPlanePointSet (MAV_window *w, MAV_class *c, MAV_callbackPlanePointFn fn) {
  mav_callbackSet (mav_callback_planePoint, w, c, (MAV_callbackFn) fn);
}

int mav_callbackPlanePointExec (MAV_window *w, MAV_object *o, MAV_line ln, MAV_vector *pp) {
  int rv= 0;
  MAV_BB bb;

  if (mav_callbackQuery (mav_callback_planePoint, w, o)) {
    rv= mav_callbackExec (mav_callback_planePoint, w, o, (void *) &ln, (void *) pp);
  } else if (mav_callbackQuery (mav_callback_BB, w, o)) {
      if (mav_callbackBBExec (w, o, &bb)) rv= mav_BBPlanePoint (bb, ln, pp);
  }

  return rv;
}


