/*
 GNU Maverik - a system for managing display and interaction in 
               Virtual Environment applications.
 Copyright (C) 1999-2002 Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - maverik@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/

#include <stdio.h>
#include <math.h>
#include "maverik.h"
#include "mav_collide.h"
#include "mav_ffnav.h"
#include "mavlib_ffnav.h"

float mav_navForcesZone= 0.4;
float mav_navForcesGradient= 0.8;
float mav_navForceRevMax= 0.1;
float mav_navForceRotMax= 0.01;

MAV_person *mavlib_forcesUser;
MAV_vector mavlib_forcesForwards;
MAV_vector mavlib_forcesVR;
float mavlib_leftForce;
float mavlib_rightForce;
float mavlib_revForce;
MAV_vector mavlib_totalForce;

void mavlib_navForceLocal (MAV_object *obj, MAV_drawInfo *di, void *params)
{
  MAV_vector pt;
  float max_dist;

  if (mav_callbackNearestPointExec (mav_win_current, obj,
		(void *) mavlib_forcesUser, (void *) &pt)) {
    MAV_vector centre;
    float hit_dist;
    MAV_vector force_vector;

    if (mav_navPlaneXY) {
      centre.x= mavlib_forcesUser->vp->eye.x + mavlib_forcesUser->view.x *
	(mavlib_forcesUser->eye_inset - mavlib_forcesUser->depth*0.5);
      centre.y= mavlib_forcesUser->vp->eye.y + mavlib_forcesUser->view.y *
	(mavlib_forcesUser->eye_inset - mavlib_forcesUser->depth*0.5);
      centre.z= 0;

      force_vector= mav_vectorSub (centre, pt);
      force_vector.z= 0;

      hit_dist= sqrt (mav_vectorDotProduct (force_vector, force_vector));
      hit_dist -= mavlib_forcesUser->width*0.5;
      if (hit_dist < 0) hit_dist= 0.0;
      force_vector= mav_vectorNormalize (force_vector);

      max_dist= sqrt (mavlib_forcesUser->width*mavlib_forcesUser->width+
	mavlib_forcesUser->depth*mavlib_forcesUser->depth)*0.5 -
	mavlib_forcesUser->width*0.5;

      if (((mav_navForcesZone - hit_dist) > 0.0) &&
	(hit_dist <= max_dist+MAV_EPSILON)) {
        force_vector.x *= mav_navForcesGradient *
		(mav_navForcesZone - hit_dist);
        force_vector.y *= mav_navForcesGradient *
		(mav_navForcesZone - hit_dist);

        mavlib_totalForce= mav_vectorAdd (mavlib_totalForce, force_vector);
      }
    } else {
      centre.x= mavlib_forcesUser->vp->eye.x + mavlib_forcesUser->view.x *
        (mavlib_forcesUser->eye_inset - mavlib_forcesUser->depth*0.5);
      centre.y= 0;
      centre.z= mavlib_forcesUser->vp->eye.z + mavlib_forcesUser->view.z *
        (mavlib_forcesUser->eye_inset - mavlib_forcesUser->depth*0.5);

      force_vector= mav_vectorSub (centre, pt);
      force_vector.y= 0;

      hit_dist= sqrt (mav_vectorDotProduct (force_vector, force_vector));
      hit_dist -= mavlib_forcesUser->width*0.5;
      if (hit_dist < 0) hit_dist= 0.0;
      force_vector= mav_vectorNormalize (force_vector);

      max_dist= sqrt (mavlib_forcesUser->width*mavlib_forcesUser->width+
        mavlib_forcesUser->depth*mavlib_forcesUser->depth)*0.5 -
        mavlib_forcesUser->width*0.5;

      if (((mav_navForcesZone - hit_dist) > 0.0) &&
        (hit_dist <= max_dist+MAV_EPSILON)) {
        force_vector.x *= mav_navForcesGradient *
                (mav_navForcesZone - hit_dist);
        force_vector.z *= mav_navForcesGradient *
                (mav_navForcesZone - hit_dist);

        mavlib_totalForce= mav_vectorAdd (mavlib_totalForce, force_vector);
      }
    }
  }
}

int mavlib_navForcesLocalSum (MAV_person *the_user, MAV_list *sms_list, MAV_vector *res_force)
{
  MAV_drawInfo di;
  MAV_SMS *sms;
  MAV_SMSExecFn sum_fn;
  MAV_vector pt;
  float force_size;

/* set shared variables for mavlib_navForceLocal */
  mavlib_totalForce.x= 0;
  mavlib_totalForce.y= 0;
  mavlib_totalForce.z= 0;

  mavlib_forcesUser= the_user;

/* set up clip planes */
  di.cp.num= 6;

  di.cp.planes[0].norm= the_user->view;
  di.cp.planes[1].norm= mav_vectorScalar (the_user->view, -1.0);
  di.cp.planes[2].norm= the_user->up;
  di.cp.planes[3].norm= mav_vectorScalar (the_user->up, -1.0);
  di.cp.planes[4].norm= the_user->right;
  di.cp.planes[5].norm= mav_vectorScalar (the_user->right, -1.0);

  pt= mav_vectorAdd (
	mav_vectorAdd (
	  the_user->vp->eye, 
	  mav_vectorScalar (the_user->up, the_user->forehead)),
	mav_vectorAdd (
	  mav_vectorScalar (the_user->right,
		the_user->width*0.5 + mav_navForcesZone),
	  mav_vectorScalar (the_user->view, 
		the_user->eye_inset + mav_navForcesZone)));

  di.cp.planes[0].d= mav_vectorDotProduct (di.cp.planes[0].norm, pt);
  di.cp.planes[2].d= mav_vectorDotProduct (di.cp.planes[2].norm, pt);
  di.cp.planes[4].d= mav_vectorDotProduct (di.cp.planes[4].norm, pt);

  pt= mav_vectorSub (
	mav_vectorSub (pt,
	  mav_vectorScalar (the_user->up, the_user->height - mav_navEpsilon)),
	mav_vectorAdd (
	  mav_vectorScalar (the_user->right,
		the_user->width + mav_navForcesZone * 2.0),
	  mav_vectorScalar (the_user->view,
		the_user->depth + mav_navForcesZone * 2.0)));

  di.cp.planes[1].d= mav_vectorDotProduct (di.cp.planes[1].norm, pt);
  di.cp.planes[3].d= mav_vectorDotProduct (di.cp.planes[3].norm, pt);
  di.cp.planes[5].d= mav_vectorDotProduct (di.cp.planes[5].norm, pt);

  sum_fn.fn= mavlib_navForceLocal;

/* step through list culling each SMS */
  mav_listPointerReset (sms_list);
  while (mav_listItemNext (sms_list, (void *) &sms))
    mav_SMSCallbackExecFnExec (sms, &di, &sum_fn);

  force_size= mav_vectorDotProduct (mavlib_totalForce, mavlib_totalForce);

  if (force_size > 0.0) {
    force_size= sqrt (force_size);
    if (force_size > mav_navForceRotMax) {
      force_size= mav_navForceRotMax;
    }

    mavlib_totalForce= mav_vectorNormalize (mavlib_totalForce);

    *res_force= mav_vectorScalar (mavlib_totalForce, force_size);

    return 1;
  }

  return 0;
}

void mavlib_navObjectForce (MAV_object *obj, MAV_drawInfo *di, void *params)
{
  float curr_rev, curr_sideways;
  MAV_vector pt, force_vector;
  float hit_dist;
  MAV_vector centre;

  if (mav_callbackNearestPointExec (mav_win_current, obj, (void *) mavlib_forcesUser, (void *) &pt)) {
    if (mav_navPlaneXY) {
      centre= mav_vectorAdd (mavlib_forcesUser->vp->eye,
	mav_vectorScalar (mavlib_forcesUser->view,
	  mavlib_forcesUser->eye_inset - mavlib_forcesUser->depth * 0.5));

      force_vector= mav_vectorSub (centre, pt);
      force_vector.z= 0;
  
      hit_dist= sqrt (mav_vectorDotProduct (force_vector, force_vector));

      if (hit_dist < MAV_EPSILON) hit_dist= MAV_EPSILON;

      force_vector= mav_vectorNormalize (force_vector);

      if ((mav_navForcesZone - hit_dist) > 0) {
        force_vector.x *= mav_navForcesGradient * (mav_navForcesZone - hit_dist);
        force_vector.y *= mav_navForcesGradient * (mav_navForcesZone - hit_dist);

        curr_rev= mav_vectorDotProduct (force_vector, mavlib_forcesForwards);
        if (curr_rev < MAV_EPSILON) {
	  curr_sideways= mav_vectorDotProduct (force_vector, mavlib_forcesVR);

	  if (curr_sideways > mavlib_rightForce)
	    mavlib_rightForce= curr_sideways;
	  if (curr_sideways < mavlib_leftForce)
	    mavlib_leftForce= curr_sideways;

	  if (curr_rev < mavlib_revForce)
	    mavlib_revForce= curr_rev;
        }
      }
    } else {
      centre= mav_vectorAdd (mavlib_forcesUser->vp->eye,
        mav_vectorScalar (mavlib_forcesUser->view,
          mavlib_forcesUser->eye_inset - mavlib_forcesUser->depth * 0.5));

      force_vector= mav_vectorSub (centre, pt);
      force_vector.y= 0;
  
      hit_dist= sqrt (mav_vectorDotProduct (force_vector, force_vector));

      if (hit_dist < MAV_EPSILON) hit_dist= MAV_EPSILON;

      force_vector= mav_vectorNormalize (force_vector);

      if ((mav_navForcesZone - hit_dist) > 0) {
        force_vector.x *= mav_navForcesGradient * (mav_navForcesZone - hit_dist);
        force_vector.z *= mav_navForcesGradient * (mav_navForcesZone - hit_dist);

        curr_rev= mav_vectorDotProduct (force_vector, mavlib_forcesForwards);
        if (curr_rev < MAV_EPSILON) {
          curr_sideways= mav_vectorDotProduct (force_vector, mavlib_forcesVR);

          if (curr_sideways > mavlib_rightForce)
            mavlib_rightForce= curr_sideways;
          if (curr_sideways < mavlib_leftForce)
            mavlib_leftForce= curr_sideways;

          if (curr_rev < mavlib_revForce)
            mavlib_revForce= curr_rev;
        }
      }
    }
  }
}

void mavlib_navForcesCompute (MAV_person *the_user, MAV_list *sms_list, float *max_range, MAV_vector *dir, float original_force)
{
  MAV_drawInfo di;
  MAV_vector pt;
  MAV_SMS *sms;
  MAV_SMSExecFn forces_fn;
  MAV_vector new_dir;

/* set variables for mavlib_navObjectForce */
  mavlib_forcesUser= the_user;
  mavlib_revForce= 0;
  mavlib_leftForce= 0;
  mavlib_rightForce= 0;

  if (mav_navPlaneXY) {
    mavlib_forcesForwards= *dir;
    mavlib_forcesForwards.z= 0;
    mavlib_forcesForwards= mav_vectorNormalize (mavlib_forcesForwards);

    mavlib_forcesVR.x= mavlib_forcesForwards.y;
    mavlib_forcesVR.y= -mavlib_forcesForwards.x;
    mavlib_forcesVR.z= 0;
  } else {
    mavlib_forcesForwards= *dir;
    mavlib_forcesForwards.y= 0;
    mavlib_forcesForwards= mav_vectorNormalize (mavlib_forcesForwards);

    mavlib_forcesVR.x= -mavlib_forcesForwards.z;
    mavlib_forcesVR.y= 0;
    mavlib_forcesVR.z= mavlib_forcesForwards.x;
  }

/* set up clip planes for forcefield region */
  di.cp.num= 6;

  di.cp.planes[0].norm= mavlib_forcesForwards;
  di.cp.planes[1].norm= mav_vectorScalar (mavlib_forcesForwards, -1.0);
  di.cp.planes[2].norm= the_user->up;
  di.cp.planes[3].norm= mav_vectorScalar (the_user->up, -1.0);
  di.cp.planes[4].norm= mavlib_forcesVR;
  di.cp.planes[5].norm= mav_vectorScalar (mavlib_forcesVR, -1.0);

  pt= mav_vectorAdd (the_user->vp->eye,
	mav_vectorAdd (
	  mav_vectorScalar (the_user->up, the_user->forehead),
	  mav_vectorScalar (mavlib_forcesVR,
		mav_navForcesZone + the_user->width)));

  di.cp.planes[1].d= mav_vectorDotProduct (di.cp.planes[1].norm, pt);
  di.cp.planes[2].d= mav_vectorDotProduct (di.cp.planes[2].norm, pt);
  di.cp.planes[4].d= mav_vectorDotProduct (di.cp.planes[4].norm, pt);

  pt= mav_vectorSub (
	mav_vectorAdd (pt,
	  mav_vectorScalar (mavlib_forcesForwards,
		mav_navForcesZone + the_user->width)),
	mav_vectorAdd (
	  mav_vectorScalar (mavlib_forcesVR,
		2.0 * (mav_navForcesZone + the_user->width)),
	  mav_vectorScalar (the_user->up, the_user->height - the_user->knees)));

  di.cp.planes[0].d= mav_vectorDotProduct (di.cp.planes[0].norm, pt);
  di.cp.planes[3].d= mav_vectorDotProduct (di.cp.planes[3].norm, pt);
  di.cp.planes[5].d= mav_vectorDotProduct (di.cp.planes[5].norm, pt);

/* set up SMS exec function */
  forces_fn.fn= mavlib_navObjectForce;

/* step through list of SMS's */
  mav_listPointerReset (sms_list);
  while (mav_listItemNext (sms_list, (void *) &sms))
    mav_SMSCallbackExecFnExec (sms, &di, &forces_fn);

/* compute new direction of movement using force values */
  if (mavlib_revForce < (-mav_navForceRevMax - original_force))
    mavlib_revForce= -mav_navForceRevMax - original_force;

  if (mav_navPlaneXY) {
    new_dir.x= mavlib_forcesVR.x * (mavlib_leftForce + mavlib_rightForce) +
	mavlib_forcesForwards.x * (original_force + mavlib_revForce);
    new_dir.y= mavlib_forcesVR.y * (mavlib_leftForce + mavlib_rightForce) +
	mavlib_forcesForwards.y * (original_force + mavlib_revForce);
    new_dir.z= 0;
  } else {
    new_dir.x= mavlib_forcesVR.x * (mavlib_leftForce + mavlib_rightForce) +
        mavlib_forcesForwards.x * (original_force + mavlib_revForce);
    new_dir.y= 0;
    new_dir.z= mavlib_forcesVR.z * (mavlib_leftForce + mavlib_rightForce) +
        mavlib_forcesForwards.z * (original_force + mavlib_revForce);
  }

  *max_range= mav_vectorDotProduct (new_dir, new_dir);
  if (*max_range>0.0) *max_range= sqrt (*max_range);

  *dir= mav_vectorNormalize (new_dir);
}

