/*
 GNU Maverik - a system for managing display and interaction in 
               Virtual Environment applications.
 Copyright (C) 1999-2002 Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - maverik@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/

#include <math.h>
#include "maverik.h"
#include "mav_collide.h"
#include "mav_ffnav.h"
#include "mavlib_ffnav.h"

float col_dist;
float max_move;

int mavlib_navMoveTest (MAV_person *the_user, MAV_list *sms_list, float dist, MAV_vector dir)
{
/* get position of any face of the_user that faces in the direction of */
/* travel and do a collision check on it */
  int rv= 1;
  MAV_polygon face_poly;
  MAV_vector poly_vert[4];

  max_move= dist;

  face_poly.vert= poly_vert;
  face_poly.np= 4;
  face_poly.matrix= MAV_ID_MATRIX;

/* check which sides of user are moving forwards */

/* front face */
  if (mav_vectorDotProduct (the_user->view, dir) > MAV_EPSILON) {
/* get four corners of front of user */
    face_poly.vert[0]= mav_vectorAdd (
	mav_vectorAdd (the_user->vp->eye,
	  mav_vectorScalar (the_user->up, the_user->forehead)),
	mav_vectorAdd (
	  mav_vectorScalar (the_user->right, the_user->width * 0.5),
	  mav_vectorScalar (the_user->view, the_user->eye_inset)));

    face_poly.vert[1]= mav_vectorSub (face_poly.vert[0],
	mav_vectorScalar (the_user->up, the_user->height - mav_navEpsilon));

    face_poly.vert[2]= mav_vectorSub (face_poly.vert[1],
	mav_vectorScalar (the_user->right, the_user->width));

    face_poly.vert[3]= mav_vectorAdd (face_poly.vert[2],
	mav_vectorScalar (the_user->up, the_user->height - mav_navEpsilon));

    col_dist= dist+MAV_EPSILON;

    if (mavlib_navCollideTest (&face_poly, dir, &col_dist, sms_list)) {
      rv= 1;
      if ((col_dist - MAV_EPSILON) - max_move) max_move= col_dist-MAV_EPSILON;
    } else rv= 0;

  }

/* back face */
  if (rv) {
    if (mav_vectorDotProduct (the_user->view, dir) < -MAV_EPSILON) {
/* get four corners of back of user */
      face_poly.vert[0]= mav_vectorAdd (
	mav_vectorAdd (the_user->vp->eye,
	  mav_vectorScalar (the_user->up, the_user->forehead)),
	mav_vectorAdd (
	  mav_vectorScalar (the_user->right, the_user->width * 0.5),
	  mav_vectorScalar (the_user->view,
		 the_user->eye_inset - the_user->depth)));

      face_poly.vert[1]= mav_vectorSub (face_poly.vert[0],
	mav_vectorScalar (the_user->up, the_user->height - mav_navEpsilon));

      face_poly.vert[2]= mav_vectorSub (face_poly.vert[1],
	mav_vectorScalar (the_user->right, the_user->width));

      face_poly.vert[3]= mav_vectorAdd (face_poly.vert[2],
	mav_vectorScalar (the_user->up, the_user->height - mav_navEpsilon));

      col_dist= dist+MAV_EPSILON;
      if (mavlib_navCollideTest (&face_poly, dir, &col_dist, sms_list)) {
        rv= 1;
        if ((col_dist-MAV_EPSILON)<max_move) max_move= col_dist-MAV_EPSILON;
      } else rv= 0;
    }
  }

/* right face */
  if (rv) {
    if (mav_vectorDotProduct (the_user->right, dir) > MAV_EPSILON) {
/* get four corners of right side of user */
      face_poly.vert[0]= mav_vectorAdd (
	mav_vectorAdd (the_user->vp->eye,
	  mav_vectorScalar (the_user->up, the_user->forehead)),
	mav_vectorAdd (
	  mav_vectorScalar (the_user->right, the_user->width * 0.5),
	  mav_vectorScalar (the_user->view, the_user->eye_inset)));

      face_poly.vert[1]= mav_vectorSub (face_poly.vert[0],
	mav_vectorScalar (the_user->up, the_user->height - mav_navEpsilon));

      face_poly.vert[2]= mav_vectorSub (face_poly.vert[1],
	mav_vectorScalar (the_user->view, the_user->depth));

      face_poly.vert[3]= mav_vectorAdd (face_poly.vert[2],
	mav_vectorScalar (the_user->up, the_user->height - mav_navEpsilon));

      col_dist= dist+MAV_EPSILON;

      if (mavlib_navCollideTest (&face_poly, dir, &col_dist, sms_list)) {
        rv= 1;
        if ((col_dist-MAV_EPSILON)<max_move) max_move= col_dist-MAV_EPSILON;
      } else rv= 0;
    }
  }

/* left face */
  if (rv) {
    if (mav_vectorDotProduct (the_user->right, dir) < -MAV_EPSILON) {
/* get four corners of left side of user */
      face_poly.vert[0]= mav_vectorAdd (
	mav_vectorAdd (the_user->vp->eye,
	  mav_vectorScalar (the_user->up, the_user->forehead)),
	mav_vectorSub (
	  mav_vectorScalar (the_user->view, the_user->eye_inset),
	  mav_vectorScalar (the_user->right, the_user->width * 0.5)));

      face_poly.vert[1]= mav_vectorSub (face_poly.vert[0],
	mav_vectorScalar (the_user->up, the_user->height - mav_navEpsilon));

      face_poly.vert[2]= mav_vectorSub (face_poly.vert[1],
	mav_vectorScalar (the_user->view, the_user->depth));

      face_poly.vert[3]= mav_vectorAdd (face_poly.vert[2],
	mav_vectorScalar (the_user->up, the_user->height - mav_navEpsilon));

      col_dist= dist+MAV_EPSILON;
      if (mavlib_navCollideTest (&face_poly, dir, &col_dist, sms_list)) {
        rv= 1;
        if ((col_dist-MAV_EPSILON)<max_move) max_move= col_dist-MAV_EPSILON;
      } else rv= 0;
    }
  }

/* top face */
  if (rv) {
    if (mav_vectorDotProduct (the_user->up, dir) > MAV_EPSILON) {
/* get four corners of top of user */
      face_poly.vert[0]= mav_vectorAdd (
	mav_vectorAdd (the_user->vp->eye,
	  mav_vectorScalar (the_user->up, the_user->forehead)),
	mav_vectorAdd (
	  mav_vectorScalar (the_user->right, the_user->width * 0.5),
	  mav_vectorScalar (the_user->view, the_user->eye_inset)));

      face_poly.vert[1]= mav_vectorSub (face_poly.vert[0],
	mav_vectorScalar (the_user->view, the_user->depth));

      face_poly.vert[2]= mav_vectorSub (face_poly.vert[1],
	mav_vectorScalar (the_user->right, the_user->width));

      face_poly.vert[3]= mav_vectorAdd (face_poly.vert[2],
	mav_vectorScalar (the_user->view, the_user->depth));

      col_dist= dist+MAV_EPSILON;

      if (mavlib_navCollideTest (&face_poly, dir, &col_dist, sms_list)) {
        rv= 1;
        if ((col_dist-MAV_EPSILON)<max_move) max_move= col_dist-MAV_EPSILON;
      } else rv= 0;
    }
  }

/* bottom face */
  if (rv) {
    if (mav_vectorDotProduct (the_user->up, dir) < -MAV_EPSILON) {
/* get four corners of bottom of user */
/* use a slightly raised polygon to avoid errors */
      face_poly.vert[0]= mav_vectorAdd (
	mav_vectorAdd (the_user->vp->eye,
	  mav_vectorScalar (the_user->up,
		the_user->forehead - the_user->height + 2.0*mav_navEpsilon)),
	mav_vectorAdd (
	  mav_vectorScalar (the_user->right, the_user->width * 0.5),
	  mav_vectorScalar (the_user->view, the_user->eye_inset)));

      face_poly.vert[1]= mav_vectorSub (face_poly.vert[0],
	mav_vectorScalar (the_user->view, the_user->depth));

      face_poly.vert[2]= mav_vectorSub (face_poly.vert[1],
	mav_vectorScalar (the_user->right, the_user->width));

      face_poly.vert[3]= mav_vectorAdd (face_poly.vert[2],
	mav_vectorScalar (the_user->view, the_user->depth));

/* need longer test distance because polygon is raised slightly */
      col_dist= dist+MAV_EPSILON+2.0*mav_navEpsilon;

      if (mavlib_navCollideTest (&face_poly, dir, &col_dist, sms_list)) {
        rv= 1;
        col_dist -= 2.0*mav_navEpsilon;
        if ((col_dist-MAV_EPSILON)<max_move) max_move= col_dist-MAV_EPSILON;
      } else rv= 0;
    }
  }

  return rv;
}

void mavlib_userParamsSet (MAV_person *);

int mavlib_navFootTest (MAV_person *the_user, MAV_vector down, float dist, MAV_list *sms_list)
{
  int i;
  MAV_polygon test_poly;
  MAV_vector test_vert[4];
  int hit_something= 0;
  int not_finished= 1;
  float slab_size;
  float test_dist;
  float dist_left;

  test_poly.vert= test_vert;
  test_poly.np= 4;
  test_poly.matrix= MAV_ID_MATRIX;

  test_poly.vert[0]= mav_vectorAdd (
	mav_vectorAdd (the_user->vp->eye,
	  mav_vectorScalar (the_user->view,
		the_user->eye_inset + the_user->shoesize)),
	mav_vectorAdd (
	  mav_vectorScalar (the_user->right,
		the_user->width * 0.5 + the_user->shoesize),
	  mav_vectorScalar (the_user->up,
		the_user->forehead - the_user->height + dist)));

  test_poly.vert[1]= mav_vectorSub (test_poly.vert[0],
	mav_vectorScalar (the_user->view,
		 the_user->depth + 2.0 * the_user->shoesize));

  test_poly.vert[2]= mav_vectorSub (test_poly.vert[1],
	mav_vectorScalar (the_user->right,
		the_user->width + 2.0 * the_user->shoesize));

  test_poly.vert[3]= mav_vectorAdd (test_poly.vert[2],
	mav_vectorScalar (the_user->view,
		the_user->depth + 2.0 * the_user->shoesize));

  slab_size= dist / ((float)mav_navFootAccuracy - 1.0);

  dist_left= dist+slab_size;

  i= 0;
  while (not_finished) {
    test_dist= dist_left;
    if (mavlib_navCollideTest (&test_poly, down, &test_dist, sms_list)) {
      if (test_dist < (dist_left-MAV_EPSILON)) {
	if (mav_navPlaneXY) {
	  the_user->vp->eye.z= test_poly.vert[0].z - test_dist +
		the_user->height - the_user->forehead;
	} else {
	  the_user->vp->eye.y= test_poly.vert[0].y - test_dist +
		the_user->height - the_user->forehead;
	}

	hit_something= 1;
	not_finished= 0;
      } else
	not_finished= 0;
    }

    if (mav_navPlaneXY) {
      test_poly.vert[0].z -= slab_size;
      test_poly.vert[1].z -= slab_size;
      test_poly.vert[2].z -= slab_size;
      test_poly.vert[3].z -= slab_size;
    } else {
      test_poly.vert[0].y -= slab_size;
      test_poly.vert[1].y -= slab_size;
      test_poly.vert[2].y -= slab_size;
      test_poly.vert[3].y -= slab_size;
    }

    test_dist -= slab_size;

    i ++;
    if (i==mav_navFootAccuracy) not_finished= 0;
  }

  return hit_something;
}

