/*
 GNU Maverik - a system for managing display and interaction in 
               Virtual Environment applications.
 Copyright (C) 1999-2002 Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - maverik@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/

#include <stdio.h>
#include <math.h>
#include "maverik.h"
#include "mav_collide.h"
#include "mav_ffnav.h"
#include "mavlib_ffnav.h"

/* variables for mavlib_navPolygonTest to use */
int mavlib_navCollision;
MAV_vector mavlib_navDir;
float mavlib_navDist;
MAV_polygon *mavlib_navPoly;
MAV_vector mavlib_sideNorms[4];
float mavlib_sideResults[4];
MAV_vector mavlib_sidePlanePts[4][MAV_COLLIDE_MAX_POINTS];
int mavlib_sidePPCount[4];
MAV_clipPlanes *mavlib_di_cp;

void mavlib_navPolygonTest (MAV_object *obj, MAV_drawInfo *di, void *params)
{
  int count, i;
  MAV_line ln;
  MAV_objectIntersection oi;
  MAV_vector hit_pt;
  float old_dist= mavlib_navDist;

  if (!mavlib_navCollision) {
/* first check near polygon */
    if (mav_collidePolygonObjectPointGet (mavlib_navPoly, obj, &hit_pt)) {
      mavlib_navCollision= 1;
    } else {
/* check intersection along each edge to far side */
      ln.dir= mavlib_navDir;
      for (count= 0; count<4; count ++) {
        ln.pt= mavlib_navPoly->vert[count];

	if (mav_callbackIntersectExec (mav_win_current, obj, ln, &oi))
	  if ((oi.pt1>-MAV_EPSILON) && (oi.pt1<mavlib_navDist))
	    mavlib_navDist= oi.pt1;
      }

      if (mavlib_navDist < MAV_EPSILON) {
	mavlib_navCollision= 1;
      } else {
/* check 4 sides */
	for (count= 0; count < 4; count ++) {
	  ln.pt= mavlib_navPoly->vert[count];
	  ln.dir= mavlib_sideNorms[count];
	  mavlib_sidePPCount[count]= mav_callbackPlanePointExec (
		mav_win_current, obj, ln, mavlib_sidePlanePts[count]);

	  if (mavlib_sidePPCount[count]>0) {
	    float this;
	    int in_count;

/* check each plane point against all but far side */
	    for (i=0; i<mavlib_sidePPCount[count]; i++) {
	      in_count= 0;
	      if (mav_vectorDotProduct (mavlib_sidePlanePts[count][i],
			mavlib_sideNorms[(count+1)%4]) >=
			mavlib_sideResults[(count+1)%4])
		in_count ++;
	      if (mav_vectorDotProduct (mavlib_sidePlanePts[count][i],
			mavlib_sideNorms[(count+3)%4]) >=
			mavlib_sideResults[(count+3)%4])
		in_count ++;

/* get distance to near side (i.e. nearest that far side can be) */
	      if ((in_count==2)||(in_count==0)) {
		this= mav_vectorDotProduct (mav_vectorSub (mavlib_navPoly->vert[0],
			mavlib_sidePlanePts[count][i]), mavlib_navPoly->norm) /
			mav_vectorDotProduct (mav_vectorScalar (mavlib_navDir, -1.0),
			mavlib_navPoly->norm);

		if ((this>-MAV_EPSILON) && (this < mavlib_navDist))
		  mavlib_navDist= this;
	      }
	    } /* for (mavlib_sidePPCount) */
	  } /* if (mavlib_sidePPCount) */
	} /* for (count) */

	if (mavlib_navDist < MAV_EPSILON) {
	  mavlib_navCollision= 1;
	} else {
/* check far side */
	  MAV_polygon test_poly;
	  MAV_vector test_vert[4];
	  MAV_vector scaled_dir;
	  float last_good, curr_move, move_amount, move_offset;
	  int this_hit;

	  scaled_dir= mav_vectorScalar (mavlib_navDir, mavlib_navDist);

	  for (count=0; count<4; count++)
	    test_vert[count]= mav_vectorAdd (mavlib_navPoly->vert[count],
		scaled_dir);

	  test_poly.vert= test_vert;
	  test_poly.np= 4;
	  test_poly.matrix= MAV_ID_MATRIX;
	  test_poly.norm= mavlib_navPoly->norm;

	  move_amount= mavlib_navDist;
	  curr_move= mavlib_navDist;

	  if (mav_collidePolygonObjectPointGet (&test_poly, obj, &hit_pt)) {
	    move_amount= -mavlib_navDist*0.5;
	    move_offset= -move_amount*0.5;

/* set up line for inside test */
	    mav_callbackGetPointExec (mav_win_current, obj, &(ln.pt));
	    ln.dir= mavlib_navDir;

	    last_good= 0;
	    for (i=0; i<mav_navCollisionAccuracy; i++) {
	      if (move_offset < mav_navMoveMin) i= mav_navCollisionAccuracy;

	      for (count=0; count<4; count++) {
		scaled_dir= mav_vectorScalar (mavlib_navDir, move_amount);
	        test_vert[count]= mav_vectorAdd (test_vert[count],
			scaled_dir);
	      }
	      curr_move += move_amount;

/* test a ray from obj */

	      this_hit= 0;
	      if (mav_linePolygonIntersection (&test_poly, ln, &oi)) {
		if ((oi.pt1 > -MAV_EPSILON) && (oi.pt1 <= curr_move))
		  this_hit= 1;

	      }

	      if (this_hit || mav_collidePolygonObjectPointGet (&test_poly, obj, &hit_pt)) {
		move_amount= -move_offset;
	      } else {
		last_good= curr_move;
		move_amount= move_offset;
	      }
	      move_offset *= 0.5;
	    } /* for (i<mav_navCollisionAccuracy) */

	    mavlib_navDist= last_good;

	    if (mavlib_navDist < MAV_EPSILON) {
	      mavlib_navCollision= 1;
	    }
	  } /* if (collide) */
	} /* else (mavlib_navDist < mav_navMoveMin) */
      } /* else (mavlib_navDist < mav_navMoveMin) */
    } /* else (mavlib_navDist < mav_navMoveMin) */

/* check if far plane can be moved nearer */
    if (mavlib_navDist < old_dist) {
      mavlib_di_cp->planes[0].d -= (old_dist-mavlib_navDist) *
		mav_vectorDotProduct (mavlib_navDir,
			mavlib_di_cp->planes[0].norm);
    }
      
  } /* if (!mavlib_navCollision) */
}


int mavlib_navCollideTest (MAV_polygon *poly, MAV_vector dir, float *dist, MAV_list *sms_list)
{
/* NOTE - assumes a 4 sided polygon */
  MAV_SMS *sms;
  MAV_drawInfo di;
  MAV_SMSExecFn cull_fn;
  MAV_vector pt;
  int rv= 0;

/* set polygon normal */
  poly->norm= mav_vectorNormalize (mav_vectorCrossProduct (
	mav_vectorSub (poly->vert[1], poly->vert[0]),
	mav_vectorSub (poly->vert[2], poly->vert[1])));

/* set variables for mavlib_navPolygonTest */
  mavlib_navPoly= poly;
  mavlib_navDir= dir;
  mavlib_navDist= *dist;
  mavlib_navCollision= 0;
  mavlib_sideNorms[0]= mav_vectorNormalize (mav_vectorCrossProduct (
	mav_vectorSub (poly->vert[1], poly->vert[0]), dir));
  mavlib_sideNorms[1]= mav_vectorNormalize (mav_vectorCrossProduct (
	mav_vectorSub (poly->vert[2], poly->vert[1]), dir));
  mavlib_sideNorms[2]= mav_vectorNormalize (mav_vectorCrossProduct (
	mav_vectorSub (poly->vert[3], poly->vert[2]), dir));
  mavlib_sideNorms[3]= mav_vectorNormalize (mav_vectorCrossProduct (
	mav_vectorSub (poly->vert[0], poly->vert[3]), dir));

  mavlib_sideResults[0]= mav_vectorDotProduct (poly->vert[0],
		mavlib_sideNorms[0]);
  mavlib_sideResults[1]= mav_vectorDotProduct (poly->vert[1],
		mavlib_sideNorms[1]);
  mavlib_sideResults[2]= mav_vectorDotProduct (poly->vert[2],
		mavlib_sideNorms[2]);
  mavlib_sideResults[3]= mav_vectorDotProduct (poly->vert[3],
		mavlib_sideNorms[3]);

/* compute clip planes for the region poly will move through */
  di.cp.planes[0].norm= mav_vectorCrossProduct (
        mav_vectorSub (poly->vert[1], poly->vert[0]),
        mav_vectorSub (poly->vert[2], poly->vert[1]));

/* get a point on opposite side to poly */
  pt= mav_vectorAdd (poly->vert[0], mav_vectorScalar (dir, *dist));

/* set clip plane normals (depends on normal direction) */
  if (mav_vectorDotProduct (di.cp.planes[0].norm, dir)<0.0) {
    di.cp.planes[1].norm= di.cp.planes[0].norm;
    di.cp.planes[0].norm= mav_vectorScalar (di.cp.planes[0].norm, -1.0);
    di.cp.planes[2].norm= mav_vectorCrossProduct (
        dir, mav_vectorSub (poly->vert[1], poly->vert[0]));
    di.cp.planes[3].norm= mav_vectorCrossProduct (
        dir, mav_vectorSub (poly->vert[3], poly->vert[2]));
    di.cp.planes[4].norm= mav_vectorCrossProduct (
        dir, mav_vectorSub (poly->vert[2], poly->vert[1]));
    di.cp.planes[5].norm= mav_vectorCrossProduct (
        dir, mav_vectorSub (poly->vert[0], poly->vert[3]));
  } else {
    di.cp.planes[1].norm= mav_vectorScalar (di.cp.planes[0].norm, -1.0);
    di.cp.planes[2].norm= mav_vectorCrossProduct (
        mav_vectorSub (poly->vert[1], poly->vert[0]), dir);
    di.cp.planes[3].norm= mav_vectorCrossProduct (
        mav_vectorSub (poly->vert[3], poly->vert[2]), dir);
    di.cp.planes[4].norm= mav_vectorCrossProduct (
        mav_vectorSub (poly->vert[2], poly->vert[1]), dir);
    di.cp.planes[5].norm= mav_vectorCrossProduct (
        mav_vectorSub (poly->vert[0], poly->vert[3]), dir);
  }

  di.cp.planes[0].norm= mav_vectorNormalize (di.cp.planes[0].norm);
  di.cp.planes[1].norm= mav_vectorNormalize (di.cp.planes[1].norm);
  di.cp.planes[2].norm= mav_vectorNormalize (di.cp.planes[2].norm);
  di.cp.planes[3].norm= mav_vectorNormalize (di.cp.planes[3].norm);
  di.cp.planes[4].norm= mav_vectorNormalize (di.cp.planes[4].norm);
  di.cp.planes[5].norm= mav_vectorNormalize (di.cp.planes[5].norm);

  di.cp.planes[0].d= mav_vectorDotProduct (di.cp.planes[0].norm, pt);
  di.cp.planes[1].d= mav_vectorDotProduct (di.cp.planes[1].norm,
        poly->vert[0]);
  di.cp.planes[2].d= mav_vectorDotProduct (di.cp.planes[2].norm,
        poly->vert[0]);
  di.cp.planes[3].d= mav_vectorDotProduct (di.cp.planes[3].norm,
        poly->vert[2]);
  di.cp.planes[4].d= mav_vectorDotProduct (di.cp.planes[4].norm,
        poly->vert[1]);
  di.cp.planes[5].d= mav_vectorDotProduct (di.cp.planes[5].norm,
        poly->vert[0]);

/* set up global clip plane */

  mavlib_di_cp= &di.cp;

  di.cp.num= 6;

/* set SMSExecFn.fn */
  cull_fn.fn= mavlib_navPolygonTest;

/* step through sms_list culling each SMS */
  mav_listPointerReset (sms_list);
  while (mav_listItemNext (sms_list, (void *) &sms))
    mav_SMSCallbackExecFnExec (sms, &di, &cull_fn);

/* check result of collision tests */
  if (!mavlib_navCollision) {
    rv= 1;
    *dist= mavlib_navDist;
  }

  return rv;
}


