/*
 GNU Maverik - a system for managing display and interaction in 
               Virtual Environment applications.
 Copyright (C) 1999-2002 Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - maverik@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/

#include <stdio.h>
#include <math.h>
#include "maverik.h"
#include "mav_collide.h"
#include "mav_ffnav.h"
#include "mavlib_ffnav.h"

/**********************************************/
/* navigation with forcefields and collisions */
/**********************************************/

extern float max_move;

int mavlib_navInitialized= MAV_FALSE;
int mav_navTimer= 0;

float mav_navEpsilon= 0.01;
float mav_navFallMax= 0.32;
float mav_navClimbMax= 0.3;
float mav_navMoveMax= 100;
float mav_navMoveMin= 0.0001;

int mav_navPlaneXY= 0;
int mav_navGravity= 0;
int mav_navCollisions= 0;
int mav_navFoot= 0;
int mav_navForces= 0;
int mav_navCollisionAccuracy= 8;
int mav_navFootAccuracy= 4;

void mavlib_userParamsSet (MAV_person *the_user)
{
  if (mav_navPlaneXY) {
    the_user->view= the_user->vp->view;
    the_user->view.z= 0;
    the_user->view= mav_vectorNormalize (the_user->view);

    the_user->up.x= 0;
    the_user->up.y= 0;
    the_user->up.z= 1;

    the_user->right= the_user->vp->right;
    the_user->right.z= 0;
    the_user->right= mav_vectorNormalize (the_user->right);
  } else {
    the_user->view= the_user->vp->view;
    the_user->view.y= 0;
    the_user->view= mav_vectorNormalize (the_user->view);

    the_user->up.x= 0;
    the_user->up.y= 1;
    the_user->up.z= 0;

    the_user->right= the_user->vp->right;
    the_user->right.y= 0;
    the_user->right= mav_vectorNormalize (the_user->right);
  }
}

MAV_polygon mavlib_polyFront;
int mavlib_rotateCollision;
float mavlib_testDist;
MAV_vector mavlib_testDir;

void mavlib_rotateObjTest (MAV_object *obj, MAV_drawInfo *di, void *params)
{
  if (!mavlib_rotateCollision)
    if (mav_collidePolygonObjectVectorTest (&mavlib_polyFront, obj,
	mavlib_testDir, mavlib_testDist))
      mavlib_rotateCollision= 1;
}

void mav_navRotate (MAV_person *the_user, float angle, MAV_list *sms_list)
{
  MAV_vector new_view;
  MAV_vector new_right;
  MAV_vector rot_force;
  MAV_timer rot_timer;

  if (mav_navTimer) mav_timerStart (&rot_timer);

  mavlib_userParamsSet (the_user);

  new_view= mav_vectorRotate (the_user->vp->view, the_user->vp->fixed_up,
	angle);
  new_right= mav_vectorRotate (the_user->vp->right, the_user->vp->fixed_up,
	angle);

  if (mav_navCollisions) {
    MAV_vector new_user_view;
    MAV_vector new_user_right;
    MAV_vector poly_vert[4];
    MAV_drawInfo di;
    MAV_vector pt;
    MAV_SMSExecFn rotate_fn;
    MAV_SMS *sms;

    if (mav_navPlaneXY) {
      new_user_view= new_view;
      new_user_view.z= 0;
      new_user_view= mav_vectorNormalize (new_user_view);

      new_user_right= new_right;
      new_user_right.z= 0;
      new_user_right= mav_vectorNormalize (new_user_right);
    } else {
      new_user_view= new_view;
      new_user_view.y= 0;
      new_user_view= mav_vectorNormalize (new_user_view);

      new_user_right= new_right;
      new_user_right.y= 0;
      new_user_right= mav_vectorNormalize (new_user_right);
    }

    mavlib_testDir= mav_vectorScalar (new_user_view, -1.0);


/* now check this rotation doesn't cause a collision */
    mavlib_polyFront.vert= poly_vert;

/* get four corners of front of user */
    poly_vert[0]= mav_vectorAdd (
	mav_vectorAdd (the_user->vp->eye,
	  mav_vectorScalar (the_user->up, the_user->forehead)),
	mav_vectorAdd (
	  mav_vectorScalar (new_user_right, the_user->width * 0.5),
	  mav_vectorScalar (new_user_view, the_user->eye_inset)));

    poly_vert[1]= mav_vectorSub (poly_vert[0],
	mav_vectorScalar (the_user->up, the_user->height - mav_navEpsilon));

    poly_vert[2]= mav_vectorSub (poly_vert[1],
	mav_vectorScalar (new_user_right, the_user->width));

    poly_vert[3]= mav_vectorAdd (poly_vert[2],
	mav_vectorScalar (the_user->up, the_user->height - mav_navEpsilon));

    mavlib_polyFront.np= 4;

/* compute clip planes for user */
    di.cp.planes[0].norm= new_user_view;
    di.cp.planes[1].norm= mav_vectorScalar (new_user_view, -1.0);
    di.cp.planes[2].norm= the_user->up;
    di.cp.planes[3].norm= mav_vectorScalar (the_user->up, -1.0);
    di.cp.planes[4].norm= new_user_right;
    di.cp.planes[5].norm= mav_vectorScalar (new_user_right, -1.0);

    di.cp.planes[0].d= mav_vectorDotProduct (di.cp.planes[0].norm,
        poly_vert[0]);

    pt= mav_vectorAdd (poly_vert[0],
	  mav_vectorScalar (mavlib_testDir, the_user->depth));

    di.cp.planes[1].d= mav_vectorDotProduct (di.cp.planes[1].norm, pt);
    if (mav_navPlaneXY) {
      di.cp.planes[2].d= poly_vert[0].z;
      di.cp.planes[3].d= -1.0 * poly_vert[1].z;
    } else {
      di.cp.planes[2].d= poly_vert[0].y;
      di.cp.planes[3].d= -1.0 * poly_vert[1].y;
    }
    di.cp.planes[4].d= mav_vectorDotProduct (di.cp.planes[4].norm,
        poly_vert[0]);
    di.cp.planes[5].d= mav_vectorDotProduct (di.cp.planes[5].norm,
        poly_vert[2]);

    di.cp.num= 6;

    rotate_fn.fn= mavlib_rotateObjTest;

    mavlib_testDist= the_user->depth;

/* reset collision flag */
    mavlib_rotateCollision= 0;

/* step through sms list */
    mav_listPointerReset (sms_list);
    while (mav_listItemNext (sms_list, (void *) &sms))
      mav_SMSCallbackExecFnExec (sms, &di, &rotate_fn);

/* check for collision */
    if (!mavlib_rotateCollision) {
      the_user->vp->view= new_view;
      the_user->vp->right= new_right;
      the_user->vp->up= mav_vectorRotate (the_user->vp->up,
	the_user->vp->fixed_up, angle);
    }
  } else {
    the_user->vp->view= new_view;
    the_user->vp->right= new_right;
    the_user->vp->up= mav_vectorRotate (the_user->vp->up,
	the_user->vp->fixed_up, angle);
  }

  mavlib_userParamsSet (the_user);

  if (mav_navForces)
/* compute all local forces */
    if (mavlib_navForcesLocalSum (the_user, sms_list, &rot_force))
/* check forces don't move user into a collision */
      if (mavlib_navMoveTest (the_user, sms_list, 1.0, rot_force))
	the_user->vp->eye= mav_vectorAdd (the_user->vp->eye,
		mav_vectorScalar (rot_force, max_move));

  if (mav_navTimer) {
    mav_timerStop (&rot_timer);
    mav_timerPrint ("Rot:", rot_timer);
  }
}

void mav_navDrop (MAV_person *the_user, MAV_list *sms_list)
{
  MAV_vector down;
  float top_height, start_height;
  float foot_pos=0;
  MAV_line ln;
  MAV_SMS *sms;
  MAV_objectIntersection oi;
  float curr_drop=0;
  int int_hit= 0;
  MAV_object *dummy_obj;
  int foot_hit= 0;
  MAV_timer drop_timer;
  float fall_height;

  if (mav_navTimer) mav_timerStart (&drop_timer);

  if (mav_navGravity) {
    mavlib_userParamsSet (the_user);

    down= mav_vectorScalar (the_user->up, -1.0);

    if (mav_navPlaneXY)
      start_height= the_user->vp->eye.z;
    else
      start_height= the_user->vp->eye.y;

    if (mav_navFoot) {
/* check if user can step up */
/* first raise user by "knees" - this tests how much head clearance */
/* there is for stepping up */
      mav_navMove (the_user, the_user->up, the_user->knees, sms_list);



/* store this height and reset to original position */
      if (mav_navPlaneXY) {
	top_height= the_user->vp->eye.z;
        the_user->vp->eye.z= start_height;
      } else {
	top_height= the_user->vp->eye.y;
	the_user->vp->eye.y= start_height;
      }


/* check foot */
      if (mavlib_navFootTest (the_user, down, top_height - start_height, sms_list)) {
/* store this height and reset again */
	if (mav_navPlaneXY) {
	  foot_pos= the_user->vp->eye.z;
	  the_user->vp->eye.z= start_height;
	} else {
	  foot_pos= the_user->vp->eye.y;
	  the_user->vp->eye.y= start_height;
	}


	foot_hit= 1;
      }
    }

/* cast a ray down */
    ln.pt= the_user->vp->eye;
    if (mav_navPlaneXY)
      ln.pt.z += the_user->forehead - the_user->height*0.5;
    else
      ln.pt.y += the_user->forehead - the_user->height*0.5;

    ln.dir= down;
    mav_listPointerReset (sms_list);
    while (mav_listItemNext (sms_list, (void *) &sms))
      if (mav_SMSCallbackIntersectExec (sms, mav_win_current, ln, &oi, &dummy_obj))
	if ((!int_hit)||(int_hit && (oi.pt1 < curr_drop))) {
	  curr_drop= oi.pt1;
	  int_hit= 1;
	}


/* drop user */
    mav_navMove (the_user, down, mav_navFallMax, sms_list);


/* if foot position is higher then user then use that */
    if (mav_navPlaneXY) {
      if (foot_hit && (foot_pos > the_user->vp->eye.z))
        the_user->vp->eye.z= foot_pos;
    } else {
      if (foot_hit && (foot_pos > the_user->vp->eye.y))
	the_user->vp->eye.y= foot_pos;
    }


/* if line returned intersection check if it is higher than user */
    if (int_hit) {
      curr_drop= start_height - curr_drop + the_user->height*0.5 - the_user->forehead;
      if (mav_navPlaneXY) {
        if (curr_drop > the_user->vp->eye.z)
	  the_user->vp->eye.z= curr_drop;
      } else {
	if (curr_drop > the_user->vp->eye.y)
	  the_user->vp->eye.y= curr_drop;
        }
    }


/* set "falling" flag */
    if (mav_navPlaneXY)
      fall_height= start_height - the_user->vp->eye.z;
    else
      fall_height= start_height - the_user->vp->eye.y;

    if ((fall_height>0.0) && (fall_height<mav_navEpsilon)) {
      if (mav_navPlaneXY)
	the_user->vp->eye.z= start_height;
      else
	the_user->vp->eye.y= start_height;
    }

    if (fall_height >= mav_navFallMax*0.5)
      the_user->falling= MAV_TRUE;
    else the_user->falling= MAV_FALSE;
  } else the_user->falling= MAV_FALSE;

  if (mav_navTimer) {
    mav_timerStop (&drop_timer);
    mav_timerPrint ("Drop:", drop_timer);
  }
}

void mav_navMove (MAV_person *the_user, MAV_vector dir, float dist, MAV_list *sms_list)
{
  mavlib_userParamsSet (the_user);

  if (dist > MAV_EPSILON) {
    if (mav_navCollisions) {
/* fix if negative dist */
      if (dist < 0) {
        dir= mav_vectorScalar (dir, -1.0);
        dist= -dist;
      }

      if (mavlib_navMoveTest (the_user, sms_list, dist, dir)) {
        the_user->vp->eye= mav_vectorAdd (the_user->vp->eye,
		mav_vectorScalar (dir, max_move-MAV_EPSILON));
      }
    } else {
/* move without collision tests */
      the_user->vp->eye= mav_vectorAdd (the_user->vp->eye,
        mav_vectorScalar (dir, dist));
    }
  }
}

void mav_navForcesMove (MAV_person *the_user, MAV_vector dir, float dist, MAV_list *sms_list)
{
  float max;
  MAV_vector old_eye= the_user->vp->eye;
  float divisor;
  MAV_timer move_timer;

  if (mav_navTimer) mav_timerStart (&move_timer);

/* some maths to cope with a non-unit direction vector or negative dist */
  divisor= mav_vectorDotProduct (dir, dir);
  divisor= sqrt (divisor);
  dir= mav_vectorNormalize (dir);
  dist *= divisor;

  if (dist < 0) {
    dir= mav_vectorScalar (dir, -1.0);
    dist= -dist;
  }

/* move without forces */
  mav_navMove (the_user, dir, dist, sms_list);

  if (mav_navForces) {
/* do forces calculation here */
    mavlib_navForcesCompute (the_user, sms_list, &max, &dir, dist);

/* reset position */
    the_user->vp->eye= old_eye;

    divisor= mav_vectorDotProduct (dir, dir);
    divisor= sqrt (divisor);
    dir= mav_vectorNormalize (dir);
    max *= divisor;

    if (max < 0) {
      dir= mav_vectorScalar (dir, -1.0);
      max= -max;
    }

/* move with forces */
    mav_navMove (the_user, dir, max, sms_list);
  }

  if (mav_navTimer) {
    mav_timerStop (&move_timer);
    mav_timerPrint ("Move:", move_timer);
  }
}

char *mav_navModuleID (void) {
  return "Forcefield navigation";
}

void mav_navInit (void)
{
  if (!mavlib_navInitialized) {
    mav_collideModuleInit ();
    mav_nearestPointInit ();

    mavlib_navInitialized= MAV_TRUE;

/* add this module */
    mav_moduleNew (mav_navModuleID);
  }
}

