/*
 GNU Maverik - a system for managing display and interaction in 
               Virtual Environment applications.
 Copyright (C) 1999-2002 Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - maverik@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/

#include <stdio.h>
#include <math.h>
#include "maverik.h"
#include "mav_ffnav.h"
#include "mav_nearestPoint.h"

MAV_callback *mav_callback_nearestPoint;

/* BB nearest point function */
int mav_BBNearestPoint (MAV_BB bb, MAV_person *the_user, MAV_vector *pt) {
  float user_top, user_bottom;
  MAV_vector centre;

  if (mav_navPlaneXY) {
    centre.x= the_user->vp->eye.x + the_user->view.x * (the_user->eye_inset-the_user->depth*0.5);
    centre.y= the_user->vp->eye.y + the_user->view.y * (the_user->eye_inset-the_user->depth*0.5);
    centre.z= the_user->vp->eye.z;

    user_top= centre.z + the_user->forehead;
    user_bottom= user_top - the_user->height + the_user->knees + MAV_EPSILON;

  /* first check bb at user height */
    if ((bb.max.z < user_bottom) || (bb.min.z > user_top))
      return 0;

    *pt= centre;

  /* now check the distance in x/y plane to nearest pt on bb */

  /* get distance to nearest y side */
    if (centre.y < bb.min.y)
      pt->y= bb.min.y;
    else if (centre.y > bb.max.y)
      pt->y= bb.max.y;

  /* get distance to nearest x side */
    if (centre.x < bb.min.x)
      pt->x= bb.min.x;
    else if (centre.x > bb.max.x)
      pt->x= bb.max.x;
  } else {
    centre.x= the_user->vp->eye.x + the_user->view.x * (the_user->eye_inset-the_user->depth*0.5);
    centre.y= the_user->vp->eye.y;
    centre.z= the_user->vp->eye.z + the_user->view.z * (the_user->eye_inset-the_user->depth*0.5);

    user_top= centre.y + the_user->forehead;
    user_bottom= user_top - the_user->height + the_user->knees + MAV_EPSILON;

  /* first check bb at user height */
    if ((bb.max.y < user_bottom) || (bb.min.y > user_top))
      return 0;

    *pt= centre;

  /* now check the distance in z/x plane to nearest pt on bb */

  /* get distance to nearest z side */
    if (centre.z < bb.min.z)
      pt->z= bb.min.z;
    else if (centre.z > bb.max.z)
      pt->z= bb.max.z;

  /* get distance to nearest x side */
    if (centre.x < bb.min.x)
      pt->x= bb.min.x;
    else if (centre.x > bb.max.x)
      pt->x= bb.max.x;
  }

  return 1;
}

void mav_lineNearestPoint (float, float, float, float, float, float, float, float, float, float *, MAV_vector *);
int nearest_get_next_point (MAV_person *, MAV_vector *, int *, int, MAV_vector,
MAV_vector *);

/* function to get nearest point to a vertex list */

void mav_vertsNearestPoint (MAV_person *the_user, MAV_vector *verts, int nv, float *current, MAV_vector *pt)
{
  int status;
  int index= 0;
  int extra_edge= 0;
  MAV_vector curr_pt, next_pt;

  status= nearest_get_next_point (the_user, verts, &index, nv, verts[0], &curr_pt);
  if (status) {
    if (status==2) {
      extra_edge= 1;
    }

    while (nearest_get_next_point (the_user, verts, &index, nv, curr_pt,
		&next_pt)) {
      mav_lineNearestPoint (the_user->vp->eye.x, the_user->vp->eye.y,
		the_user->vp->eye.z,
		curr_pt.x, curr_pt.y, curr_pt.z, next_pt.x, next_pt.y,
		next_pt.z, current, pt);
      curr_pt= next_pt;
    }

    if (extra_edge) {
      mav_lineNearestPoint (the_user->vp->eye.x, the_user->vp->eye.y,
		the_user->vp->eye.z,
		curr_pt.x, curr_pt.y, curr_pt.z, next_pt.x, next_pt.y,
		next_pt.z, current, pt);
    }

  }
}

/* callback functions for each object type */

/* box */

int mav_boxNearestPoint (MAV_object *obj, MAV_person *the_user, MAV_vector *pt)
{
  MAV_box *box= (MAV_box *) mav_objectDataGet (obj);
  MAV_vector vert_list[4];
  float current= MAV_INFINITY;
  MAV_vector vert[8];
  float w,h,d;
  int index;

  w= box->size.x*0.5;
  h= box->size.y*0.5;
  d= box->size.z*0.5;

  vert[0].x = -w;
  vert[0].y = -h;
  vert[0].z = -d;

  vert[1].x = -w;
  vert[1].y = h;
  vert[1].z = -d;

  vert[2].x = w;
  vert[2].y = -h;
  vert[2].z = -d;

  vert[3].x = w;
  vert[3].y = h;
  vert[3].z = -d;

  vert[4].x = -w;
  vert[4].y = -h;
  vert[4].z = d;

  vert[5].x = -w;
  vert[5].y = h;
  vert[5].z = d;

  vert[6].x = w;
  vert[6].y = -h;
  vert[6].z = d;

  vert[7].x = w;
  vert[7].y = h;
  vert[7].z = d;

  for (index=0; index<8; index++)
    vert[index]= mav_vectorMult (vert[index], box->matrix);

/* test each face in turn */
/* face 0 */
  vert_list[0]= vert[0];
  vert_list[1]= vert[1];
  vert_list[2]= vert[3];
  vert_list[3]= vert[2];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 1 */
  vert_list[0]= vert[2];
  vert_list[1]= vert[3];
  vert_list[2]= vert[7];
  vert_list[3]= vert[6];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 2 */
  vert_list[0]= vert[1];
  vert_list[1]= vert[3];
  vert_list[2]= vert[7];
  vert_list[3]= vert[5];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 3 */
  vert_list[0]= vert[0];
  vert_list[1]= vert[1];
  vert_list[2]= vert[5];
  vert_list[3]= vert[4];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 4 */
  vert_list[0]= vert[0];
  vert_list[1]= vert[2];
  vert_list[2]= vert[6];
  vert_list[3]= vert[4];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 5 */
  vert_list[0]= vert[4];
  vert_list[1]= vert[5];
  vert_list[2]= vert[7];
  vert_list[3]= vert[6];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);


  if (current==MAV_INFINITY) return 0;

  return 1;
}

/* cone */

int mav_coneNearestPoint (MAV_object *obj, MAV_person *the_user, MAV_vector *pt) {
  MAV_cone *cone= (MAV_cone *) mav_objectDataGet (obj);
  MAV_vector vert_list[4];
  float current= MAV_INFINITY;
  MAV_vector vert[8];
  float w,d,h;
  int index;

  w= cone->rb;
  h= cone->rb;
  d= cone->height*0.5;

  vert[0].x = -w;
  vert[0].y = -h;
  vert[0].z = -d;

  vert[1].x = -w;
  vert[1].y = h;
  vert[1].z = -d;

  vert[2].x = w;
  vert[2].y = -h;
  vert[2].z = -d;

  vert[3].x = w;
  vert[3].y = h;
  vert[3].z = -d;

  w= cone->rt;
  h= cone->rt;

  vert[4].x = -w;
  vert[4].y = -h;
  vert[4].z = d;

  vert[5].x = -w;
  vert[5].y = h;
  vert[5].z = d;

  vert[6].x = w;
  vert[6].y = -h;
  vert[6].z = d;

  vert[7].x = w;
  vert[7].y = h;
  vert[7].z = d;

  for (index=0; index<8; index++)
    vert[index]= mav_vectorMult (vert[index], cone->matrix);

/* test each face in turn */
/* face 0 */
  vert_list[0]= vert[0];
  vert_list[1]= vert[1];
  vert_list[2]= vert[3];
  vert_list[3]= vert[2];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 1 */
  vert_list[0]= vert[2];
  vert_list[1]= vert[3];
  vert_list[2]= vert[7];
  vert_list[3]= vert[6];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 2 */
  vert_list[0]= vert[1];
  vert_list[1]= vert[3];
  vert_list[2]= vert[7];
  vert_list[3]= vert[5];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 3 */
  vert_list[0]= vert[0];
  vert_list[1]= vert[1];
  vert_list[2]= vert[5];
  vert_list[3]= vert[4];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 4 */
  vert_list[0]= vert[0];
  vert_list[1]= vert[2];
  vert_list[2]= vert[6];
  vert_list[3]= vert[4];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 5 */
  vert_list[0]= vert[4];
  vert_list[1]= vert[5];
  vert_list[2]= vert[7];
  vert_list[3]= vert[6];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);


  if (current==MAV_INFINITY) return 0;

  return 1;
}

/* circular torus */

int mav_ctorusNearestPoint (MAV_object *obj, MAV_person *the_user, MAV_vector *pt) {
  /*  MAV_ctorus *ct= (MAV_ctorus *) mav_objectDataGet (obj);*/

  return 1;
}

/* cylinder */

int mav_cylinderNearestPoint (MAV_object *obj, MAV_person *the_user, MAV_vector *pt) {
  MAV_cylinder *cyl= (MAV_cylinder *) mav_objectDataGet (obj);
  MAV_vector vert_list[4];
  float current= MAV_INFINITY;
  MAV_vector vert[8];
  float w,d,h;
  int index;

  w= cyl->radius;
  h= cyl->radius;
  d= cyl->height*0.5;

  vert[0].x = -w;
  vert[0].y = -h;
  vert[0].z = -d;

  vert[1].x = -w;
  vert[1].y = h;
  vert[1].z = -d;

  vert[2].x = w;
  vert[2].y = -h;
  vert[2].z = -d;

  vert[3].x = w;
  vert[3].y = h;
  vert[3].z = -d;

  vert[4].x = -w;
  vert[4].y = -h;
  vert[4].z = d;

  vert[5].x = -w;
  vert[5].y = h;
  vert[5].z = d;

  vert[6].x = w;
  vert[6].y = -h;
  vert[6].z = d;

  vert[7].x = w;
  vert[7].y = h;
  vert[7].z = d;

  for (index=0; index<8; index++)
    vert[index]= mav_vectorMult (vert[index], cyl->matrix);

/* test each face in turn */
/* face 0 */
  vert_list[0]= vert[0];
  vert_list[1]= vert[1];
  vert_list[2]= vert[3];
  vert_list[3]= vert[2];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 1 */
  vert_list[0]= vert[2];
  vert_list[1]= vert[3];
  vert_list[2]= vert[7];
  vert_list[3]= vert[6];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 2 */
  vert_list[0]= vert[1];
  vert_list[1]= vert[3];
  vert_list[2]= vert[7];
  vert_list[3]= vert[5];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 3 */
  vert_list[0]= vert[0];
  vert_list[1]= vert[1];
  vert_list[2]= vert[5];
  vert_list[3]= vert[4];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 4 */
  vert_list[0]= vert[0];
  vert_list[1]= vert[2];
  vert_list[2]= vert[6];
  vert_list[3]= vert[4];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 5 */
  vert_list[0]= vert[4];
  vert_list[1]= vert[5];
  vert_list[1]= vert[7];
  vert_list[1]= vert[6];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);


  if (current==MAV_INFINITY) return 0;

  return 1;
}

/* ellipse */

int mav_ellipseNearestPoint (MAV_object *obj, MAV_person *the_user, MAV_vector *pt) {
  /*  MAV_ellipse *elip= (MAV_ellipse *) mav_objectDataGet (obj);*/

  return 1;
}

/* nearest point to 2d line segment */
void mav_lineNearestPoint (float x, float y, float z, float x1, float y1, float z1, float x2, float y2, float z2, float *current, MAV_vector *point)
{
  float d1, d2, d1_to_2, dline, dnearest;

  if (mav_navPlaneXY) {
    d1= sqrt ((x-x1)*(x-x1) + (y-y1)*(y-y1));
    d1_to_2= sqrt ((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1));
  } else {
    d1= sqrt ((x-x1)*(x-x1) + (z-z1)*(z-z1));
    d1_to_2= sqrt ((x2-x1)*(x2-x1) + (z2-z1)*(z2-z1));
  }

  if (d1_to_2>0) {
    if (mav_navPlaneXY)
      dline= (x-x1)*(x2-x1)/d1_to_2 + (y-y1)*(y2-y1)/d1_to_2;
    else
      dline= (x-x1)*(x2-x1)/d1_to_2 + (z-z1)*(z2-z1)/d1_to_2;
  } else dline= -1.0;

  if (dline>0 && dline<d1_to_2) {
    dnearest= sqrt (d1*d1 - dline*dline);
    if (dnearest< *current) {
      *current= dnearest;
      point->x= x1 + dline * (x2-x1)/d1_to_2;
      point->y= y1 + dline * (y2-y1)/d1_to_2; 
      point->z= z1 + dline * (z2-z1)/d1_to_2;
    }
  } else {
    if (mav_navPlaneXY)
      d2= sqrt ((x-x2)*(x-x2) + (y-y2)*(y-y2));
    else
      d2= sqrt ((x-x2)*(x-x2) + (z-z2)*(z-z2));

    if (d1< *current) {
      *current= d1;
      point->x= x1;
      if (mav_navPlaneXY) {
	point->y= y1;
	point->z= z1 + dline * (z2-z1)/d1_to_2;
      } else {
	point->y= y1 + dline * (y2-y1)/d1_to_2;
	point->z= z1;
      }
    }

    if (d2< *current) {
      *current= d2;
      point->x= x2;
      if (mav_navPlaneXY) {
	point->y= y2;
	point->z= z1 + dline * (z2-z1)/d1_to_2;
      } else {
	point->y= y1 + dline * (y2-y1)/d1_to_2;
	point->z= z1;
      }
    }
  }
}

int nearest_get_next_point (MAV_person *the_user, MAV_vector *verts, int *curr, int maxverts, MAV_vector last_pt, MAV_vector *next_pt)
{
  int rv= 0;
  int not_finished;
  float top, bot;

  if ((*curr)==(maxverts+1)) return 0;

  if (mav_navPlaneXY)
    top= the_user->vp->eye.z + the_user->forehead;
  else
    top= the_user->vp->eye.y + the_user->forehead;

  bot= top - the_user->height + the_user->knees;

  if ((mav_navPlaneXY && last_pt.z < (bot - MAV_EPSILON)) ||
	(!mav_navPlaneXY && last_pt.y < (bot - MAV_EPSILON))) {
/* check for bottom plane intersection */
    not_finished= 1;
    while (not_finished) {
      if ((mav_navPlaneXY && verts[(*curr)%maxverts].z > bot) ||
	  (!mav_navPlaneXY && verts[(*curr)%maxverts].y > bot)) {
	not_finished= 0;
	rv= 2;
	if (mav_navPlaneXY) {
	  next_pt->x= last_pt.x + (verts[(*curr)%maxverts].x-last_pt.x)*
		(bot-last_pt.z)/(verts[(*curr)%maxverts].z-last_pt.z);
	  next_pt->y= last_pt.y + (verts[(*curr)%maxverts].y-last_pt.y)*
		(bot-last_pt.z)/(verts[(*curr)%maxverts].z-last_pt.z);
	  next_pt->z= bot;
	} else {
          next_pt->x= last_pt.x + (verts[(*curr)%maxverts].x-last_pt.x)*
                (bot-last_pt.y)/(verts[(*curr)%maxverts].y-last_pt.y);
	  next_pt->y= bot;
	  next_pt->z= last_pt.z + (verts[(*curr)%maxverts].z-last_pt.z)*
		(bot-last_pt.y)/(verts[(*curr)%maxverts].y-last_pt.y);
	}
      } else {
	last_pt= verts[(*curr)%maxverts];
	(*curr) ++;
	if ((*curr)==(maxverts+1)) {
	  not_finished= 0;
	  rv= 0;
	}
      }
    }
  } else if ((mav_navPlaneXY && last_pt.z > (top + MAV_EPSILON)) ||
		(!mav_navPlaneXY && last_pt.y > (top + MAV_EPSILON))) {
/* check for top plane intersection */
    not_finished= 1;
    while (not_finished) {
      if ((mav_navPlaneXY && verts[(*curr)%maxverts].z < top) ||
	   (!mav_navPlaneXY && verts[(*curr)%maxverts].y < top)) {
        not_finished= 0;
        rv= 2;
	if (mav_navPlaneXY) {
          next_pt->x= last_pt.x + (verts[(*curr)%maxverts].x-last_pt.x)*
                (top-last_pt.z)/(verts[(*curr)%maxverts].z-last_pt.z);
          next_pt->y= last_pt.y + (verts[(*curr)%maxverts].y-last_pt.y)*
                (top-last_pt.z)/(verts[(*curr)%maxverts].z-last_pt.z);
          next_pt->z= top;
	} else {
          next_pt->x= last_pt.x + (verts[(*curr)%maxverts].x-last_pt.x)*
                (top-last_pt.y)/(verts[(*curr)%maxverts].y-last_pt.y);
	  next_pt->y= top;
          next_pt->z= last_pt.z + (verts[(*curr)%maxverts].z-last_pt.z)*
                (top-last_pt.y)/(verts[(*curr)%maxverts].y-last_pt.y);
	}
      } else {
        last_pt= verts[(*curr)%maxverts];
        (*curr) ++;
        if ((*curr)==(maxverts+1)) {
          not_finished= 0;
          rv= 0;
        } 
      }
    }
  } else {
/* check for either plane intersection */
    if ((mav_navPlaneXY && verts[(*curr)%maxverts].z > (top + MAV_EPSILON)) ||
	(!mav_navPlaneXY && verts[(*curr)%maxverts].y > (top + MAV_EPSILON))) {
/* crosses top plane - check if we were already at a cross point */
      if ((mav_navPlaneXY && last_pt.z > (top - MAV_EPSILON)) ||
	  (!mav_navPlaneXY && last_pt.y > (top - MAV_EPSILON))) {
	last_pt= verts[(*curr)%maxverts];
	(*curr) ++;

	rv= nearest_get_next_point (the_user, verts, curr, maxverts, last_pt, next_pt);
      } else {
	rv= 1;
	if (mav_navPlaneXY) {
          next_pt->x= last_pt.x + (verts[(*curr)%maxverts].x-last_pt.x)*
                (top-last_pt.z)/(verts[(*curr)%maxverts].z-last_pt.z);
          next_pt->y= last_pt.y + (verts[(*curr)%maxverts].y-last_pt.y)*
                (top-last_pt.z)/(verts[(*curr)%maxverts].z-last_pt.z);
          next_pt->z= top;
	} else {
          next_pt->x= last_pt.x + (verts[(*curr)%maxverts].x-last_pt.x)*
                (top-last_pt.y)/(verts[(*curr)%maxverts].y-last_pt.y);
          next_pt->y= top;
	  next_pt->z= last_pt.z + (verts[(*curr)%maxverts].z-last_pt.z)*
                (top-last_pt.y)/(verts[(*curr)%maxverts].y-last_pt.y);
	}
      }
    } else if ((mav_navPlaneXY && verts[(*curr)%maxverts].z < (bot - MAV_EPSILON)) ||
	(!mav_navPlaneXY && verts[(*curr)%maxverts].y < (bot - MAV_EPSILON))) {
/* crosses bottom plane - check if we were already at a cross point */
      if ((mav_navPlaneXY && last_pt.z < (bot + MAV_EPSILON)) ||
	  (!mav_navPlaneXY && last_pt.y < (bot + MAV_EPSILON))) {
	last_pt= verts[(*curr)%maxverts];
	(*curr) ++;

	rv= nearest_get_next_point (the_user, verts, curr, maxverts, last_pt, next_pt);
      } else {
	rv= 1;
	if (mav_navPlaneXY) {
          next_pt->x= last_pt.x + (verts[(*curr)%maxverts].x-last_pt.x)*
                (bot-last_pt.z)/(verts[(*curr)%maxverts].z-last_pt.z);
          next_pt->y= last_pt.y + (verts[(*curr)%maxverts].y-last_pt.y)*
                (bot-last_pt.z)/(verts[(*curr)%maxverts].z-last_pt.z);
          next_pt->z= bot;
	} else {
          next_pt->x= last_pt.x + (verts[(*curr)%maxverts].x-last_pt.x)*
                (bot-last_pt.y)/(verts[(*curr)%maxverts].y-last_pt.y);
          next_pt->y= bot;
	  next_pt->z= last_pt.z + (verts[(*curr)%maxverts].z-last_pt.z)*
                (bot-last_pt.y)/(verts[(*curr)%maxverts].y-last_pt.y);
	}
      }
    } else {
/* doesn't cross plane */
      rv= 1;
      *next_pt= verts[(*curr)%maxverts];
      (*curr) ++;
    }
  }

  return rv;
}

/* facet group */

int mav_facetNearestPoint (MAV_object *obj, MAV_person *the_user, MAV_vector *pt) {
  MAV_facet *facet= (MAV_facet *) mav_objectDataGet (obj);
  int i, count;
  float current;
  int rv= 1;
  MAV_vector *verts;

  current= MAV_INFINITY;

  for (i=0; i<facet->npolys; i++) {
    verts= mav_malloc (facet->np[i] * sizeof (MAV_vector));
    for (count=0; count<facet->np[i]; count ++)
      verts[count]= mav_vectorMult (facet->vert[i][count], facet->matrix);

    mav_vertsNearestPoint (the_user, verts, facet->np[i], &current, pt);

    mav_free (verts);
  }

  if (current==MAV_INFINITY) return 0;

  return rv;
}

/* half ellipse */

int mav_hellipseNearestPoint (MAV_object *obj, MAV_person *the_user, MAV_vector *pt) {
  MAV_hellipse *helip= (MAV_hellipse *) mav_objectDataGet (obj);
  MAV_vector vert_list[4];
  float current= MAV_INFINITY;
  MAV_vector vert[8];
  float w,h,d;
  int index;

  w= helip->radius;
  h= helip->radius;
  d= helip->height;

  vert[0].x = -w;
  vert[0].y = -h;
  vert[0].z = 0;

  vert[1].x = -w;
  vert[1].y = h;
  vert[1].z = 0;

  vert[2].x = w;
  vert[2].y = -h;
  vert[2].z = 0;

  vert[3].x = w;
  vert[3].y = h;
  vert[3].z = 0;

  vert[4].x = -w;
  vert[4].y = -h;
  vert[4].z = d;

  vert[5].x = -w;
  vert[5].y = h;
  vert[5].z = d;

  vert[6].x = w;
  vert[6].y = -h;
  vert[6].z = d;

  vert[7].x = w;
  vert[7].y = h;
  vert[7].z = d;

  for (index=0; index<8; index++)
    vert[index]= mav_vectorMult (vert[index], helip->matrix);

/* test each face in turn */
/* face 0 */
  vert_list[0]= vert[0];
  vert_list[1]= vert[1];
  vert_list[2]= vert[3];
  vert_list[3]= vert[2];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 1 */
  vert_list[0]= vert[2];
  vert_list[1]= vert[3];
  vert_list[2]= vert[7];
  vert_list[3]= vert[6];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 2 */
  vert_list[0]= vert[1];
  vert_list[1]= vert[3];
  vert_list[2]= vert[7];
  vert_list[3]= vert[5];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 3 */
  vert_list[0]= vert[0];
  vert_list[1]= vert[1];
  vert_list[2]= vert[5];
  vert_list[3]= vert[4];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 4 */
  vert_list[0]= vert[0];
  vert_list[1]= vert[2];
  vert_list[2]= vert[6];
  vert_list[3]= vert[4];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 5 */
  vert_list[0]= vert[4];
  vert_list[1]= vert[5];
  vert_list[2]= vert[7];
  vert_list[3]= vert[6];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);


  if (current==MAV_INFINITY) return 0;

  return 1;
}

/* half sphere */

int mav_hsphereNearestPoint (MAV_object *obj, MAV_person *the_user, MAV_vector *pt) {
  /*  MAV_hsphere *hsph= (MAV_hsphere *) mav_objectDataGet (obj);*/

  return 1;
}

/* polygon */

int mav_polygonNearestPoint (MAV_object *obj, MAV_person *the_user, MAV_vector *pt) {
  MAV_polygon *poly= (MAV_polygon *) mav_objectDataGet (obj);
  float current= MAV_INFINITY;
  MAV_vector *verts;
  int count;

  verts= (MAV_vector *) mav_malloc (poly->np * sizeof (MAV_vector));
  for (count= 0; count < poly->np; count ++)
    verts[count]= mav_vectorMult (poly->vert[count], poly->matrix);

  mav_vertsNearestPoint (the_user, verts, poly->np, &current, pt);

  mav_free (verts);

  if (current==MAV_INFINITY) return 0;

  return 1;
}

/* pyramid */

int mav_pyramidNearestPoint (MAV_object *obj, MAV_person *the_user, MAV_vector *pt) {
  MAV_pyramid *pyr= (MAV_pyramid *) mav_objectDataGet (obj);
  float w, h, d, bsx, bsy, tsx, tsy;
  MAV_vector vert[8];
  MAV_vector vert_list[4];
  float current= MAV_INFINITY;
  int index;

  w= pyr->offset_x*0.5;
  h= pyr->offset_y*0.5;
  d= pyr->height*0.5;
  bsx= pyr->bot_size_x*0.5;
  bsy= pyr->bot_size_y*0.5;
  tsx= pyr->top_size_x*0.5;
  tsy= pyr->top_size_y*0.5;

/* Store the 8 vertices of the pyramid */

  vert[0].x = -w-bsx;
  vert[0].y = -h-bsy;
  vert[0].z = -d;

  vert[1].x = vert[0].x;
  vert[1].y = -h+bsy;
  vert[1].z = -d;

  vert[2].x = -w+bsx;
  vert[2].y = vert[0].y;
  vert[2].z = -d;

  vert[3].x = vert[2].x;
  vert[3].y = vert[1].y;
  vert[3].z = -d;

  vert[4].x = w-tsx;
  vert[4].y = h-tsy;
  vert[4].z = d;

  vert[5].x = vert[4].x;
  vert[5].y = h+tsy;
  vert[5].z = d;

  vert[6].x = w+tsx;
  vert[6].y = vert[4].y;
  vert[6].z = d;

  vert[7].x = vert[6].x;
  vert[7].y = vert[5].y;
  vert[7].z = d;

  for (index=0; index<8; index++)
    vert[index]= mav_vectorMult (vert[index], pyr->matrix);

/* test each face in turn */
/* face 0 */
  vert_list[0]= vert[0];
  vert_list[1]= vert[1];
  vert_list[2]= vert[3];
  vert_list[3]= vert[2];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 1 */
  vert_list[0]= vert[2];
  vert_list[1]= vert[3];
  vert_list[2]= vert[7];
  vert_list[3]= vert[6];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 2 */
  vert_list[0]= vert[1];
  vert_list[1]= vert[3];
  vert_list[2]= vert[7];
  vert_list[3]= vert[5];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 3 */
  vert_list[0]= vert[0];
  vert_list[1]= vert[1];
  vert_list[2]= vert[5];
  vert_list[3]= vert[4];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 4 */
  vert_list[0]= vert[0];
  vert_list[1]= vert[2];
  vert_list[2]= vert[6];
  vert_list[3]= vert[4];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);

/* face 5 */
  vert_list[0]= vert[4];
  vert_list[1]= vert[5];
  vert_list[2]= vert[7];
  vert_list[3]= vert[6];

  mav_vertsNearestPoint (the_user, vert_list, 4, &current, pt);


  if (current==MAV_INFINITY) return 0;

  return 1;
}

/* rectangular torus */

int mav_rtorusNearestPoint (MAV_object *obj, MAV_person *the_user, MAV_vector *pt) {
  /*  MAV_rtorus *rt= (MAV_rtorus *) mav_objectDataGet (obj);*/

  return 1;
}

/* sphere */

int mav_sphereNearestPoint (MAV_object *obj, MAV_person *the_user, MAV_vector *pt) {
  /*  MAV_sphere *sph= (MAV_sphere *) mav_objectDataGet (obj);*/

  return 1;
}

/********************************************************************/
/********************************************************************/

void mav_nearestPointInit (void) {
  mav_callback_nearestPoint= mav_callbackNew ();

  mav_callbackNearestPointSet (mav_win_all, mav_class_facet, mav_facetNearestPoint);
  mav_callbackNearestPointSet (mav_win_all, mav_class_box, mav_boxNearestPoint);
  mav_callbackNearestPointSet (mav_win_all, mav_class_pyramid, mav_pyramidNearestPoint);
  mav_callbackNearestPointSet (mav_win_all, mav_class_cylinder, mav_cylinderNearestPoint);
  mav_callbackNearestPointSet (mav_win_all, mav_class_cone, mav_coneNearestPoint);
  mav_callbackNearestPointSet (mav_win_all, mav_class_hellipse, mav_hellipseNearestPoint);
  mav_callbackNearestPointSet (mav_win_all, mav_class_polygon, mav_polygonNearestPoint);

#if 0
  mav_callbackNearestPointSet (mav_win_all, mav_class_cone, mav_coneNearestPoint);
  mav_callbackNearestPointSet (mav_win_all, mav_class_ctorus, mav_ctorusNearestPoint);
  mav_callbackNearestPointSet (mav_win_all, mav_class_cylinder, mav_cylinderNearestPoint);
  mav_callbackNearestPointSet (mav_win_all, mav_class_ellipse, mav_ellipseNearestPoint);
  mav_callbackNearestPointSet (mav_win_all, mav_class_facet, mav_facetNearestPoint);
  mav_callbackNearestPointSet (mav_win_all, mav_class_hellipse, mav_hellipseNearestPoint);
  mav_callbackNearestPointSet (mav_win_all, mav_class_hsphere, mav_hsphereNearestPoint);
  mav_callbackNearestPointSet (mav_win_all, mav_class_pyramid, mav_pyramidNearestPoint);
  mav_callbackNearestPointSet (mav_win_all, mav_class_rtorus, mav_rtorusNearestPoint);
  mav_callbackNearestPointSet (mav_win_all, mav_class_sphere, mav_sphereNearestPoint);
  mav_callbackNearestPointSet (mav_win_all, mav_class_Polygon, mav_polygonNearestPoint);
#endif
}

/* wrapper routines to set and execute get point callback */

void mav_callbackNearestPointSet (MAV_window *w, MAV_class *c, MAV_callbackNearestPointFn fn) {
  mav_callbackSet (mav_callback_nearestPoint, w, c, (MAV_callbackFn) fn);
}

int mav_callbackNearestPointExec (MAV_window *w, MAV_object *o, MAV_person *the_user, MAV_vector *pt) {
  int rv= 0;
  MAV_BB bb;

  if (mav_callbackQuery (mav_callback_nearestPoint, w, o)) {
    rv= mav_callbackExec (mav_callback_nearestPoint, w, o, (void *) the_user, (void *) pt);
  } else if (mav_callbackQuery (mav_callback_BB, w, o)) {
    if (mav_callbackBBExec (w, o, &bb)) rv= mav_BBNearestPoint (bb, the_user, pt);
  }

  return rv;
}

