/*
 GNU Maverik - a system for managing display and interaction in 
               Virtual Environment applications.
 Copyright (C) 1999-2002 Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - maverik@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/

#include <stdio.h>
#include "maverik.h"
#include "mav_ffnav.h"

/*********************************************/
/*  these must be defined in your executable */
/*********************************************/

extern MAV_person the_user;
extern MAV_list *sms_list;

/*********************************************/

#define MAV_DISTANCE_PER_SECOND 1.0
#define MAV_MAX_DISTANCE_PER_FRAME 0.1
#define MAV_ANGLE_PER_SECOND (MAV_PI/4.0)
#define MAV_MAX_ANGLE_PER_FRAME (MAV_PI/50.0)
#define MAV_RUNNING_FACTOR 3.0
#define MAV_RUNNING_ANGLE 2.0

float nav_speed= 5.0;
float nav_angle_speed= 5.0;

/* framerate factors */
float get_frame_factor (void) {
  float frame_factor;

  frame_factor= MAV_DISTANCE_PER_SECOND/mav_fps;
  if (frame_factor > MAV_MAX_DISTANCE_PER_FRAME)
    frame_factor= MAV_MAX_DISTANCE_PER_FRAME;

  return frame_factor;
}

float get_frame_angle (void) {
  float frame_angle;

  frame_angle= MAV_ANGLE_PER_SECOND/mav_fps;
  if (frame_angle > MAV_MAX_ANGLE_PER_FRAME)
    frame_angle= MAV_MAX_ANGLE_PER_FRAME;

  return frame_angle;
}

/* 2D mouse forwards and turning */
int nav_x, nav_y;
void do_nav (void *ignored) {
  float xdiff, ydiff;
  MAV_vector dir;

  if (!the_user.falling) {
    xdiff= (nav_x-mav_mouse_x)*0.5*0.005 * nav_angle_speed*get_frame_angle ();
    ydiff= (nav_y-mav_mouse_y)*0.5*0.1 * nav_speed * get_frame_factor ();

    mav_navRotate (&the_user, xdiff, sms_list);

    if (ydiff > 0.0)
      dir= the_user.view;
    else {
      dir= mav_vectorScalar (the_user.view, -1.0);
      ydiff= -ydiff;
    }

    if (mav_navPlaneXY)
      dir.z= 0;
    else
      dir.y= 0;

    dir= mav_vectorNormalize (dir);

    if (ydiff > 0.0)
      mav_navForcesMove (&the_user, dir, ydiff, sms_list);
  }
}

int nav_start (MAV_object *obj, MAV_mouseEvent *me) {
  if (me->movement==MAV_PRESSED) {
    nav_x= mav_mouse_x;
    nav_y= mav_mouse_y;
    mav_frameFn1Add (do_nav, NULL);
  } else
    mav_frameFn1Rmv (do_nav, NULL);

  return 1;
}

/* 2D mouse sideways */
int sideways_x;

void do_sideways (void *ignored) {
  float xdiff;
  MAV_vector dir;

  if (!the_user.falling) {
    xdiff= (sideways_x-mav_mouse_x)*0.5*0.1* nav_speed * get_frame_factor ();

    if (xdiff < 0.0) {
      dir= the_user.right;
      xdiff= -xdiff;
    } else
      dir= mav_vectorScalar (the_user.right, -1.0);

    if (mav_navPlaneXY)
      dir.z= 0;
    else
      dir.y= 0;

    dir= mav_vectorNormalize (dir);

    if (xdiff > 0.0)
      mav_navForcesMove (&the_user, dir, xdiff, sms_list);
  }
}

int sideways_start (MAV_object *obj, MAV_mouseEvent *me) {
  if (me->movement==MAV_PRESSED) {
    sideways_x= mav_mouse_x;
    mav_frameFn1Add (do_sideways, NULL);
  } else
    mav_frameFn1Rmv (do_sideways, NULL);

  return 1;
}


/* 2D mouse up/down */
int up_y;
int up_x;

void do_up (void *ignored) {
  float ydiff;
  float xdiff;
  MAV_vector dir;

  if (!mav_navGravity) {
    ydiff= (up_y-mav_mouse_y)*0.5*0.1 * nav_speed * get_frame_factor ();
    xdiff= (up_x-mav_mouse_x)*0.5*0.1 * nav_speed * get_frame_factor ();

    if (xdiff< 0.0) {
      dir= the_user.right;
      xdiff= -xdiff;
    } else
      dir= mav_vectorScalar (the_user.right, -1.0);

    if (mav_navPlaneXY)
      dir.z= 0;
    else
      dir.y= 0;

    dir= mav_vectorNormalize (dir);

    if (xdiff>0.0)
      mav_navForcesMove (&the_user, dir, xdiff, sms_list);

    if (ydiff>0.0)
      dir= the_user.up;
    else {
      dir= mav_vectorScalar (the_user.up, -1.0);
      ydiff= -ydiff;
    }

    if (ydiff>0.0)
      mav_navForcesMove (&the_user, dir, ydiff, sms_list);

  }
}

int up_start (MAV_object *obj, MAV_mouseEvent *me) {
  if (me->movement==MAV_PRESSED) {
    up_x= mav_mouse_x;
    up_y= mav_mouse_y;
    mav_frameFn1Add (do_up, NULL);
  } else
    mav_frameFn1Rmv (do_up, NULL);

  return 1;
}

/* keyboard navigation functions */

/* forwards */
void key_forwards (void *ignored) {
  if (!the_user.falling) {
    MAV_vector dir;

    dir= the_user.vp->view;
    if (mav_navPlaneXY)
      dir.z= 0;
    else
      dir.y= 0;

    dir= mav_vectorNormalize (dir);

    mav_navForcesMove (&the_user, dir, nav_speed * get_frame_factor (),
		sms_list);
  }
}

/* backwards */
void key_backwards (void *ignored) {
  if (!the_user.falling) {
    MAV_vector dir;

    dir= the_user.vp->view;
    if (mav_navPlaneXY)
      dir.z= 0;
    else
      dir.y= 0;

    dir= mav_vectorNormalize (dir);

    mav_navForcesMove (&the_user, dir, -nav_speed * get_frame_factor (),
		sms_list);
  }
}

/* left */
void key_left (void *ignored) {
  if (mav_keyboardGet (MAV_KEY_ALT_L) != MAV_PRESSED)
    mav_navRotate (&the_user, nav_angle_speed*get_frame_angle (), sms_list);
  else
    if (!the_user.falling) {
      MAV_vector dir;

      dir= the_user.vp->right;
      if (mav_navPlaneXY)
        dir.z= 0;
      else
	dir.y= 0;

      dir= mav_vectorNormalize (dir);

      mav_navForcesMove (&the_user, dir, -nav_speed*get_frame_factor (),
		sms_list);

    }
}

/* right */
void key_right (void *ignored) {
  if (mav_keyboardGet (MAV_KEY_ALT_L) != MAV_PRESSED)
    mav_navRotate (&the_user, -nav_angle_speed*get_frame_angle (), sms_list);
  else
    if (!the_user.falling) {
      MAV_vector dir;

      dir= the_user.vp->right;
      if (mav_navPlaneXY)
        dir.z= 0;
      else
	dir.y= 0;

      dir= mav_vectorNormalize (dir);

      mav_navForcesMove (&the_user, dir, nav_speed * get_frame_factor (),
		sms_list);
    }
}

/* pan up */
void key_pan_up (void *ignored) {
  if (mav_navPlaneXY) {
    if (the_user.vp->view.z < 0.9) {
      the_user.vp->view= mav_vectorRotate (the_user.vp->view, the_user.vp->right,
		nav_angle_speed * get_frame_angle ());
      the_user.vp->up= mav_vectorRotate (the_user.vp->up, the_user.vp->right,
		nav_angle_speed * get_frame_angle ());
    }
  } else {
    if (the_user.vp->view.y < 0.9) {
      the_user.vp->view= mav_vectorRotate (the_user.vp->view, the_user.vp->right,
                nav_angle_speed * get_frame_angle ());
      the_user.vp->up= mav_vectorRotate (the_user.vp->up, the_user.vp->right,
                nav_angle_speed * get_frame_angle ());
    }
  }
}

/* pan down */
void key_pan_down (void *ignored) {
  if (mav_navPlaneXY) {
    if (the_user.vp->view.z > -0.9) {
      the_user.vp->view= mav_vectorRotate (the_user.vp->view, the_user.vp->right,
		-nav_angle_speed * get_frame_angle ());
      the_user.vp->up= mav_vectorRotate (the_user.vp->up, the_user.vp->right,
		-nav_angle_speed * get_frame_angle ());
    }
  } else {
    if (the_user.vp->view.y > -0.9) {
      the_user.vp->view= mav_vectorRotate (the_user.vp->view, the_user.vp->right,
                -nav_angle_speed * get_frame_angle ());
      the_user.vp->up= mav_vectorRotate (the_user.vp->up, the_user.vp->right,
                -nav_angle_speed * get_frame_angle ());
    }
  }
}

/* keyboard control function */

void nav_keybb (int key, int movement) {
  switch (key) {
    case 'k':
	if (movement==MAV_PRESSED) {
fprintf (stderr, "\nkeys:\n====\n\n");
fprintf (stderr, "arrows : movement\n");
fprintf (stderr, "shift : run\n");
fprintf (stderr, "alt : strafe\n");
fprintf (stderr, "page up/down : look up/down\n");
fprintf (stderr, "end : reset view up\n");
fprintf (stderr, "\nc : toggle collisions\n");
fprintf (stderr, "g : toggle gravity\n");
fprintf (stderr, "u : auto step-up\n");
fprintf (stderr, "f : toggle forcefields\n");
fprintf (stderr, "\n[ ] : forcefield gradient (strength) change\n");
fprintf (stderr, "shift-[ ] : forcefield zone (distance) change\n");
	}
	break;
/* keyboard navigation keys */
    case MAV_KEY_UP:
	if (movement==MAV_PRESSED) {
	  mav_frameFn0Add (key_forwards, NULL);
	} else {
	  mav_frameFn0Rmv (key_forwards, NULL);
	}
	break;
    case MAV_KEY_DOWN:
	if (movement==MAV_PRESSED) {
	  mav_frameFn0Add (key_backwards, NULL);
	} else {
	  mav_frameFn0Rmv (key_backwards, NULL);
	}
	break;
    case MAV_KEY_LEFT:
	if (movement==MAV_PRESSED) {
	  mav_frameFn0Add (key_left, NULL);
	} else {
	  mav_frameFn0Rmv (key_left, NULL);
	}
	break;
    case MAV_KEY_RIGHT:
	if (movement==MAV_PRESSED) {
	  mav_frameFn0Add (key_right, NULL);
	} else {
	  mav_frameFn0Rmv (key_right, NULL);
	}
	break;
    case MAV_KEY_PAGE_UP:
	if (movement==MAV_PRESSED)
	  mav_frameFn1Add (key_pan_up, NULL);
	else
	  mav_frameFn1Rmv (key_pan_up, NULL);
	break;
    case MAV_KEY_PAGE_DOWN:
	if (movement==MAV_PRESSED)
	  mav_frameFn1Add (key_pan_down, NULL);
	else
	  mav_frameFn1Rmv (key_pan_down, NULL);
	break;
    case MAV_KEY_END:
/* reset to horizontal view direction */
	if (movement==MAV_PRESSED) {
	  if (mav_navPlaneXY)
	    the_user.vp->view.z= 0;
	  else
	    the_user.vp->view.y= 0;

	  the_user.vp->view= mav_vectorNormalize (the_user.vp->view);
	  the_user.vp->up= the_user.up;
	}
	break;
    case MAV_KEY_SHIFT_L:
    case MAV_KEY_SHIFT_R:
	if (movement==MAV_PRESSED) {
	  nav_speed*= MAV_RUNNING_FACTOR;
	  nav_angle_speed*= MAV_RUNNING_ANGLE;
	} else {
	  nav_speed= 5.0;
	  nav_angle_speed= 5.0;
	}
	break;
/* collisions, forcefield, gravity, etc. */
    case 'g':
	if (movement==MAV_PRESSED) {
	  mav_navGravity= !mav_navGravity;
	  if (mav_navGravity)
	    fprintf (stderr, "Gravity on\n");
	  else
	    fprintf (stderr, "Gravity off\n");
	}
	break;
    case 'c':
	if (movement==MAV_PRESSED) {
	  mav_navCollisions= !mav_navCollisions;
	  if (mav_navCollisions)
	    fprintf (stderr, "Collisions on\n");
	  else
	    fprintf (stderr, "Collisions off\n");
	}
	break;
    case 'u':
	if (movement==MAV_PRESSED) {
	  mav_navFoot= !mav_navFoot;
	  if (mav_navFoot)
	    fprintf (stderr, "Foot on\n");
	  else
	    fprintf (stderr, "Foot off\n");
	}
	break;
    case 'f':
	if (movement==MAV_PRESSED) {
	  mav_navForces= !mav_navForces;
	  if (mav_navForces)
	    fprintf (stderr, "Forces on\n");
	  else
	    fprintf (stderr, "Forces off\n");
	}
	break;
    case ']':
	if (movement==MAV_PRESSED) {
	  mav_navForcesZone *= 1.1;
	  fprintf (stderr, "forces zone: %f\n", mav_navForcesZone);
	}
	break;
    case '[':
	if (movement==MAV_PRESSED) {
	  mav_navForcesZone *= 0.9;
	  if (mav_navForcesZone < 0.025) mav_navForcesZone= 0.025;
	  fprintf (stderr, "forces zone: %f\n", mav_navForcesZone);
	}
	break;
    case '}':
	if (movement==MAV_PRESSED) {
	  mav_navForcesGradient += 0.1;
	  fprintf (stderr, "forces gradient: %f\n", mav_navForcesGradient);
	}
	break;
    case '{':
	if (movement==MAV_PRESSED) {
	  mav_navForcesGradient -= 0.1;
	  if (mav_navForcesGradient < 0.1) mav_navForcesGradient= 0.1;
	  fprintf (stderr, "forces gradient: %f\n", mav_navForcesGradient);
	}
	break;
    default:
	break;
  }
}


