/*
 GNU Maverik - a system for managing display and interaction in 
               Virtual Environment applications.
 Copyright (C) 1999-2002 Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - maverik@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Kilburn Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/


#include "maverik.h"
#include "mav_ffnav.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

MAV_list *sms_list;
MAV_person the_user;



MAV_object *obj;
int drawbb=0;
int yaw=0;
int yawv=0;

void nav_keybb(int key, int movement);

int keyb(MAV_object *o, MAV_keyboardEvent *ke)
{
  nav_keybb(ke->key, ke->movement);

  return 1;
}

void split_facet(MAV_SMS *sms, MAV_object *obj) 
{
  MAV_facet *facet= (MAV_facet *) mav_objectDataGet(obj);
  MAV_facet *f1;
  int polys_left= facet->npolys;
  int *np_ptr= facet->np;
  MAV_vector **norm_ptr= facet->norm;
  MAV_texCoord **tex_ptr= facet->tex;
  MAV_vector **vert_ptr= facet->vert;
  MAV_surfaceParams **sp_ptr= facet->sp;

  while (polys_left > 1) {
    f1= (MAV_facet *) mav_malloc(sizeof(MAV_facet));
    f1->npolys= 1;
    f1->np= np_ptr;
    f1->norm= norm_ptr;
    f1->tex= tex_ptr;
    f1->vert= vert_ptr;
    f1->sp= sp_ptr;
    f1->matrix= facet->matrix;
    f1->userdef= NULL;

    polys_left -= 1;
    np_ptr += 1;
    norm_ptr += 1;
    tex_ptr += 1;
    vert_ptr += 1;
    sp_ptr += 1;

    mav_SMSCallbackObjectAddExec(sms, mav_objectNew(mav_class_facet, f1));
  }

  if (polys_left > 0) {
    f1= (MAV_facet *) mav_malloc(sizeof(MAV_facet));
    f1->npolys= polys_left;
    f1->np= np_ptr;
    f1->norm= norm_ptr;
    f1->tex= tex_ptr;
    f1->vert= vert_ptr;
    f1->sp= sp_ptr;
    f1->matrix= facet->matrix;
    f1->userdef= NULL;

    mav_SMSCallbackObjectAddExec(sms, mav_objectNew(mav_class_facet, f1));
  }
}

void split_composite(MAV_SMS *sms, MAV_composite *c) 
{
  int i;

  for (i=0; i<c->numobj; i++) {

    if (mav_objectClassGet(c->obj[i])==mav_class_facet)
    {
      split_facet(sms, c->obj[i]);
    }
    else if (mav_objectClassGet(c->obj[i])==mav_class_SMSObj)
    {
      MAV_SMSObj *smsobj= (MAV_SMSObj *) mav_objectDataGet(c->obj[i]);
      MAV_object *o;

      mav_SMSCallbackPointerResetExec(smsobj->sms);
      while (mav_SMSCallbackObjectNextExec(smsobj->sms, &o)) {
	if (mav_objectClassGet(o)==mav_class_facet) 
	{
	  split_facet(sms, o);
	}
	else
	{
	  char *id;
	  mav_callbackIDExec(mav_win_current, c->obj[i], &id);      
	  printf("Unknown type [%s] in composite, ignoring\n", id);
	}
      }
    }
    else
    {
      char *id;
      mav_callbackIDExec(mav_win_current, c->obj[i], &id);      
      printf("Unknown type [%s] in composite, ignoring\n", id);
    }
  }
}


int nav_start(MAV_object *obj, MAV_mouseEvent *me);

int main(int argc, char *argv[])
{
  MAV_SMS *sms, *sms2, *sms3;
  MAV_composite c;
  MAV_BB bb;
  MAV_window *w2;
  MAV_viewParams vp2;
  MAV_box b;

/* Initialise the Maverik system */

  mav_initialise(&argc, argv);

  if (argc!=2) {
    printf("Usage: %s <filename>\n", argv[0]);
    exit(0);
  }

/* Initilise ff navigation */

  mav_navInit();

/* Turn on all forces */

  mav_navForces=1;
  mav_navCollisions=1;
  mav_navGravity=1;
  mav_navFoot=1;

/* Various dimensions of avatar */

  the_user.vp= &mav_vp_default;
  the_user.width= 0.3; 
  the_user.depth= 0.2;
  the_user.eye_inset= 0.1;
  the_user.height= 1.7;
  the_user.forehead= 0.1;

  if (!strcmp(argv[1], "gym.ac"))
  {
    the_user.knees= 0.6;    /* How height he can step up */
    the_user.shoesize= 0.1; /* Distance his foot probes */
    mav_navForcesZone=0.15; /* How far forces extend from objects */

    mav_vp_default.eye= mav_vectorSet(-1,2,-1);
    mav_vp_default.view= mav_vectorSet(0,0,1);
  }
  else if (!strcmp(argv[1], "assault.ac"))
  {
    the_user.knees= 1.0; 
    the_user.shoesize= 0.1; 
    mav_navForcesZone=0.411;

    mav_vp_default.eye= mav_vectorSet(15.6,2.08,-2);
    mav_vp_default.view= mav_vectorSet(0,0,1);
  }
  else
  {
    printf("Unknown model - using default values which may not be appropriate\n");

    the_user.knees= 1.0;
    the_user.shoesize= 0.1;
    mav_navForcesZone=0.411;
  }

  w2= mav_windowNew(0,0,400,400,"2nd view cam", NULL);
  vp2= mav_vp_default;
  vp2.eye.z=-12;
  mav_windowViewParamsSet(w2, &vp2);

/* Read the composite */

  if (mav_compositeRead(argv[1], &c, MAV_ID_MATRIX)==MAV_FALSE) {
    printf("failed to read %s.\n", argv[1]);
    exit(1);
  }

  sms= mav_SMSObjListNew();
  mav_SMSObjectAdd(sms, mav_objectNew(mav_class_composite, &c));

/* Split composite in individual objects - used for ff calcs */

  sms2= mav_SMSObjListNew();
  split_composite(sms2, &c);

  sms3= mav_SMSHBBNew();
  mav_HBBConstructFromSMS(sms3, sms2);

  sms_list= mav_listNew();
  mav_listItemAdd(sms_list, sms3);

  b.size= mav_vectorSet(the_user.width, the_user.height, the_user.depth);
  b.sp= mav_sp_default;
  b.matrix= MAV_ID_MATRIX;

/* Use ff mouse navigation */

  mav_navigationMouse(mav_win_left, nav_start);
  mav_navigationMouse(w2, mav_navigationMouseDefault);

/* Define keyboard interaction */

  mav_callbackKeyboardSet(mav_win_all, mav_class_world, keyb);

  printf("\n\n*** press 'k' for keys ***\n\n");

/* Main loop */

  while (1) {

    mav_navDrop(&the_user, sms_list);

/* Check for, and act on, any events */

    mav_eventsCheck();

/* Do what needs to be done at the start of a frame */

    mav_frameBegin();

/* Display the SMS to all windows */

    mav_SMSDisplay(mav_win_all, sms);

/* Draw BB of object if applicable */

    if (drawbb) {
      mav_callbackBBExec(mav_win_current, obj, &bb);
      mav_BBDisplay(mav_win_all, bb);
    }

    {
      MAV_object bo;

      b.matrix= mav_matrixXYZSet(b.matrix, mav_vp_default.eye);
      b.matrix.mat[MAV_MATRIX_YCOMP]= mav_vp_default.eye.y+the_user.forehead-the_user.height/2.0;

      b.matrix= mav_matrixXAxisSet(b.matrix, mav_vp_default.right);
      b.matrix= mav_matrixYAxisSet(b.matrix, mav_vp_default.up);
      b.matrix= mav_matrixZAxisSet(b.matrix, mav_vectorScalar(mav_vp_default.view, -1));

      bo.the_data= &b;
      bo.the_class= mav_class_box;

      mav_windowSet(w2);
      mav_boxDraw(&bo, NULL);
      mav_windowSet(mav_win_left);
    }

/* Do what needs to be done at the end of a frame */

    mav_frameEnd();
  }
}


